#!/usr/bin/env python

"""Exercise routines in cim_http"""

from __future__ import absolute_import

import unittest

from pywbem import cim_http


class Parse_url(unittest.TestCase):  # pylint: disable=invalid-name
    """
    Test the parse_url() function.
    """

    def _run_single(self, url, exp_host, exp_port, exp_ssl):
        '''
        Test function for single invocation of parse_url() and
        allow_defaults attribute =True(default)
        '''

        host, port, ssl = cim_http.parse_url(url)

        self.assertEqual(host, exp_host,
                         "Unexpected host: %r, expected: %r" %
                         (host, exp_host))
        self.assertEqual(port, exp_port,
                         "Unexpected port: %r, expected: %r" %
                         (port, exp_port))
        self.assertEqual(ssl, exp_ssl,
                         "Unexpected ssl: %r, expected: %r" %
                         (ssl, exp_ssl))

    def _run_single_defaults_false(self, url, exp_host, exp_port, exp_ssl):
        '''
        Test function for single invocation of parse_url() with the default
        attribute set false
        '''

        host, port, ssl = cim_http.parse_url(url, allow_defaults=False)

        self.assertEqual(host, exp_host,
                         "Unexpected host: %r, expected: %r" %
                         (host, exp_host))
        self.assertEqual(port, exp_port,
                         "Unexpected port: %r, expected: %r" %
                         (port, exp_port))
        self.assertEqual(ssl, exp_ssl,
                         "Unexpected ssl: %r, expected: %r" %
                         (ssl, exp_ssl))

    def test_all(self):
        '''
        Run all tests for parse_url().
        '''

        # Keep these defaults in sync with those in cim_http.parse_url()
        default_port_http = 5988
        default_port_https = 5989
        default_ssl = False

        self._run_single("http://my.host.com",
                         "my.host.com",
                         default_port_http,
                         False)

        self._run_single("https://my.host.com/",
                         "my.host.com",
                         default_port_https,
                         True)

        self._run_single("my.host.com",
                         "my.host.com",
                         default_port_http,
                         default_ssl)

        self._run_single("http.com",
                         "http.com",
                         default_port_http,
                         default_ssl)

        self._run_single("http.com/",
                         "http.com",
                         default_port_http,
                         default_ssl)

        self._run_single("http.com/path.segment.com",
                         "http.com",
                         default_port_http,
                         default_ssl)

        self._run_single("http.com//path.segment.com",
                         "http.com",
                         default_port_http,
                         default_ssl)

        self._run_single("http://my.host.com:1234",
                         "my.host.com",
                         1234,
                         False)

        self._run_single("http://my.host.com:1234/",
                         "my.host.com",
                         1234,
                         False)

        self._run_single("http://my.host.com:1234/path/segment",
                         "my.host.com",
                         1234,
                         False)

        self._run_single("http://9.10.11.12:1234",
                         "9.10.11.12",
                         1234,
                         False)

        self._run_single("my.host.com:1234",
                         "my.host.com",
                         1234,
                         default_ssl)

        self._run_single("my.host.com:1234/",
                         "my.host.com",
                         1234,
                         default_ssl)

        self._run_single("9.10.11.12/",
                         "9.10.11.12",
                         default_port_http,
                         default_ssl)

        self._run_single("HTTP://my.host.com",
                         "my.host.com",
                         default_port_http,
                         False)

        self._run_single("HTTPS://my.host.com",
                         "my.host.com",
                         default_port_https,
                         True)

        self._run_single("http://[2001:db8::7348]",
                         "2001:db8::7348",
                         default_port_http,
                         False)

        self._run_single("http://[2001:db8::7348-1]",
                         "2001:db8::7348%1",
                         default_port_http,
                         False)

        self._run_single("http://[2001:db8::7348-eth1]",
                         "2001:db8::7348%eth1",
                         default_port_http,
                         False)

        self._run_single("http://[2001:db8::7348-eth1]:5900",
                         "2001:db8::7348%eth1",
                         5900,
                         False)

        # Toleration of (incorrect) IPv6 URI format supported by PyWBEM:
        # Must specify port; zone index must be specified with % if used

        self._run_single("http://2001:db8::7348:1234",
                         "2001:db8::7348",
                         1234,
                         False)

        self._run_single("http://2001:db8::7348%eth0:1234",
                         "2001:db8::7348%eth0",
                         1234,
                         False)

        self._run_single("http://2001:db8::7348%1:1234",
                         "2001:db8::7348%1",
                         1234,
                         False)

        self._run_single("https://[2001:db8::7348]/",
                         "2001:db8::7348",
                         default_port_https,
                         True)

        self._run_single("http://[2001:db8::7348]:1234",
                         "2001:db8::7348",
                         1234,
                         False)

        self._run_single("https://[::ffff.9.10.11.12]:1234/",
                         "::ffff.9.10.11.12",
                         1234,
                         True)

        self._run_single("https://[::ffff.9.10.11.12-0]:1234/",
                         "::ffff.9.10.11.12%0",
                         1234,
                         True)

        self._run_single("https://[::ffff.9.10.11.12-eth0]:1234/",
                         "::ffff.9.10.11.12%eth0",
                         1234,
                         True)

        self._run_single("[2001:db8::7348]",
                         "2001:db8::7348",
                         default_port_http,
                         default_ssl)

        self._run_single("[2001:db8::7348]/",
                         "2001:db8::7348",
                         default_port_http,
                         default_ssl)

        self._run_single("[2001:db8::7348]/-eth0",
                         "2001:db8::7348",
                         default_port_http,
                         default_ssl)

    def test_all_no_defaults(self):
        """ Test urls agains parse_url with allow_defaults=False"""

        # The following tests expect good return
        self._run_single_defaults_false("http://my.host.com:5988",
                                        "my.host.com",
                                        5988,
                                        False)

        self._run_single_defaults_false("https://my.host.com:5989",
                                        "my.host.com",
                                        5989,
                                        True)

        self._run_single("HTTP://my.host.com:50000",
                         "my.host.com",
                         50000,
                         False)

        self._run_single("HTTPS://my.host.com:49000",
                         "my.host.com",
                         49000,
                         True)

        self._run_single("http://[2001:db8::7348-eth1]:5900",
                         "2001:db8::7348%eth1",
                         5900,
                         False)

        self._run_single("https://[2001:db8::7348-eth1]:5901",
                         "2001:db8::7348%eth1",
                         5901,
                         True)

        # The following tests expect errors
        try:
            self._run_single_defaults_false("my.host.com:5988",
                                            "my.host.com:5988",
                                            5988,
                                            False)
            self.fail('No Scheme: Expecting exception')
        except ValueError:
            pass

        try:
            self._run_single_defaults_false("http://my.host.com",
                                            "my.host.com",
                                            5988,
                                            False)
            self.fail('No port: Expecting exception')
        except ValueError:
            pass

        try:
            self._run_single_defaults_false("blah://my.host.com:5988",
                                            "my.host.com",
                                            5988,
                                            False)
            self.fail('Invalid Scheme: Expecting exception')
        except ValueError:
            pass

        try:
            self._run_single_defaults_false("http://my.host.com:5a98",
                                            "my.host.com",
                                            5988,
                                            False)
            self.fail('Invalid Port: Expecting exception')
        except ValueError:
            pass

        try:
            self._run_single_defaults_false("[2001:db8::7348-eth1]:5900",
                                            "2001:db8::7348%eth1",
                                            5900,
                                            False)
            self.fail('Invalid Scheme: Expecting exception')
        except ValueError:
            pass

        try:
            self._run_single_defaults_false("://[2001:db8::7348-eth1]:5900",
                                            "2001:db8::7348%eth1",
                                            5900,
                                            False)
            self.fail('Invalid Scheme: Expecting exception')
        except ValueError:
            pass

        try:
            self._run_single_defaults_false(
                "httpsx://[2001:db8::7348-eth1]:5900",
                "2001:db8::7348%eth1",
                5900,
                False)
            self.fail('Invalid Scheme: Expecting exception')
        except ValueError:
            pass

        try:
            self._run_single_defaults_false(
                "https://[2001:db8::7348-eth1]",
                "2001:db8::7348%eth1",
                5900,
                False)
            self.fail('No Port: Expecting exception')
        except ValueError:
            pass

        try:
            self._run_single_defaults_false(
                "https://2001:db8::7348-eth1",
                "2001:db8::7348%eth1",
                5900,
                False)
            self.fail('No Port: Expecting exception')
        except ValueError:
            pass

        try:
            self._run_single_defaults_false(
                "httpsx://[2001:db8::7348-eth1]:59x0",
                "2001:db8::7348%eth1",
                5900,
                False)
            self.fail('Invalid Port: Expecting exception')
        except ValueError:
            pass


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

"""
Example of handling subscriptions and indications from a particular
provider in OpenPegasus. This example depends on a test class and
method in OpenPegasus.  It creates a server and a listener, starts
the listener and then requests the indications from the server.
It waits for a defined time for all indications to be received and
then terminates either normally if all are received or in error if
not all were received.

"""
from __future__ import print_function, absolute_import

import sys
import os
import time
import threading
import logging
import datetime
from socket import getfqdn
if sys.version_info >= (3, 0):
    from urllib.parse import urlparse
if sys.version_info < (3, 0) and sys.version_info >= (2, 5):
    from urlparse import urlparse
from pywbem import WBEMConnection, WBEMServer, WBEMListener, CIMClassName, \
                   Error, ConnectionError, Uint32, WBEMSubscriptionManager

# definition of the filter.  This is openpegasus specific and uses
# a class that generates one indication for each call of a method
TEST_CLASS = 'Test_IndicationProviderClass'
TEST_CLASS_NAMESPACE = 'test/TestProvider'
TEST_QUERY = 'SELECT * from %s' % TEST_CLASS

# global count of indications recived by the local indication processor
RECEIVED_INDICATION_COUNT = 0
COUNTER_LOCK = threading.Lock()
LISTENER = None

INDICATION_START_TIME = None
LAST_INDICATION_TIME = None
MAX_TIME_BETWEEN_INDICATIONS = None
LAST_SEQ_NUM = None

LOGFILE = 'pegasusindicationtest.log'


def consume_indication(indication, host):
    """This function gets called when an indication is received.
       Depends on logger inside listener for output
    """

    #pylint: disable=global-variable-not-assigned
    global RECEIVED_INDICATION_COUNT, LISTENER, INDICATION_START_TIME, \
           LAST_INDICATION_TIME, MAX_TIME_BETWEEN_INDICATIONS, LAST_SEQ_NUM
    # increment count.
    COUNTER_LOCK.acquire()
    if INDICATION_START_TIME is None:
        INDICATION_START_TIME = datetime.datetime.now()

    if LAST_INDICATION_TIME is not None:
        time_diff = datetime.datetime.now() - LAST_INDICATION_TIME
        if MAX_TIME_BETWEEN_INDICATIONS is None:
            MAX_TIME_BETWEEN_INDICATIONS = time_diff
        elif time_diff > MAX_TIME_BETWEEN_INDICATIONS:
            MAX_TIME_BETWEEN_INDICATIONS = time_diff
    seq_num = indication['SequenceNumber']
    if LAST_SEQ_NUM is None:
        LAST_SEQ_NUM = seq_num
    else:
        if seq_num != (LAST_SEQ_NUM + 1):
            print('Missed %s indications at %s' % (((seq_num - 1) - LAST_SEQ_NUM),
                                                   LAST_SEQ_NUM))
        LAST_SEQ_NUM = seq_num

    LAST_INDICATION_TIME = datetime.datetime.now()

    RECEIVED_INDICATION_COUNT += 1
    #sys.stdout.write('.')
    #sys.stdout.flush()
    LISTENER.logger.info("Consumed CIM indication #%s: host=%s\n%s",
                         RECEIVED_INDICATION_COUNT, host, indication.tomof())
    COUNTER_LOCK.release()


def wait_for_indications(requested_indications):
    """
    Wait for indications to be received Time depends on indication count
    assume 20 per sec. Loop looks once per sec for all indications
    received but terminates after wait_time seconds.
    In addition it looks for idle indication count.  If the count has not
    moved in 10 seconds, it breaks the wait loop.
    Returns True if wait successful and counts match
    """

    last_indication_count = 0
    counter = 0
    try:
        success = False
        wait_time = int(requested_indications / 20) + 3
        stall_ct = 40           # max time period with no reception.
        for _ in range(wait_time):
            time.sleep(1)
            if last_indication_count != RECEIVED_INDICATION_COUNT:
                last_indication_count = RECEIVED_INDICATION_COUNT
                counter = 0

            # exit loop if all indications recieved.
            if RECEIVED_INDICATION_COUNT >= requested_indications:
                elapsed_time = LAST_INDICATION_TIME - INDICATION_START_TIME
                ind_per_sec = 0
                if elapsed_time.total_seconds() != 0:
                    ind_per_sec = \
                        RECEIVED_INDICATION_COUNT / elapsed_time.total_seconds()
                print('Rcvd %s indications, time=%s; %02.f per sec. '
                      'Exit wait loop' %
                      (RECEIVED_INDICATION_COUNT, elapsed_time, ind_per_sec))
                success = True
                break
            counter += 1
            if counter > stall_ct:
                elapsed_time = LAST_INDICATION_TIME - INDICATION_START_TIME
                ind_per_sec = 0
                if elapsed_time.total_seconds() != 0:
                    ind_per_sec = \
                        RECEIVED_INDICATION_COUNT / elapsed_time.total_seconds()
                print('Nothing received for %s sec: received=%s time=%s' \
                      ' %02.f per sec' % \
                      (stall_ct, RECEIVED_INDICATION_COUNT,
                       elapsed_time.total_seconds(), ind_per_sec))
                break

    # because this can be a long loop, catch cntrl-c
    except KeyboardInterrupt:
        print('Ctrl-C terminating wait loop early')

    # If success, wait and recheck to be sure no extras received.
    if success:
        time.sleep(2)
        if RECEIVED_INDICATION_COUNT != requested_indications:
            print('ERROR. Extra indications received')
    return success

def send_request_for_indications(conn, class_name, requested_indications):
    """
    Send an invokemethod to the WBEM server to initiate the indication
    output.  This is a pegasus specific operation. Note also that the
    way Pegasus works today, often the response for this request does not
    get returned until well after the indication flow has started because
    it operates on the same thread as the response.
    """
    try:
        # Send method to pegasus server to create  required number of
        # indications. This is a pegasus specific class and method

        result = conn.InvokeMethod("SendTestIndicationsCount", class_name,
                                   [('indicationSendCount',
                                     Uint32(requested_indications))])

        if result[0] != 0:
            print('SendTestIndicationCount Method error. Nonzero return=%s' \
                  % result[0])
            return False
        return True

    except Error as er:
        print('Error: Indication Method exception %s' % er)
        return False


def run_test(svr_url, listener_host, user, password, http_listener_port, \
             https_listener_port, requested_indications, repeat_loop):
    """
        Runs a test that:
        1. Creates a server
        2. Creates a dynamic listener and starts ti
        3. Creates a filter and subscription
        4. Calls the server to execute a method that creates an indication
        5. waits for indications to be received.
        6. Removes the filter and subscription and stops the listener
    """
    if os.path.exists(LOGFILE):
        os.remove(LOGFILE)
    try:
        conn = WBEMConnection(svr_url, (user, password), no_verification=True)
        server = WBEMServer(conn)

        # Create subscription_manager here to be sure we can communicate with
        # server before Creating listener, etc.
        sub_mgr = WBEMSubscriptionManager(
            subscription_manager_id='pegasusIndicationTest')

        # Add server to subscription manager
        server_id = sub_mgr.add_server(server)
        old_filters = sub_mgr.get_all_filters(server_id)
        old_subs = sub_mgr.get_all_subscriptions(server_id)
        # TODO filter for our sub mgr
        if len(old_subs) != 0 or len(old_filters) != 0:
            sub_mgr.remove_subscriptions(server_id,
                                         [inst.path for inst in old_subs])
            for filter_ in old_filters:
                sub_mgr.remove_filter(server_id, filter_.path)

    except ConnectionError as ce:
        print('Connection Error %s with %s' % (ce, svr_url))
        sys.exit(2)

    except Error as er:
        print('Error communicationg with WBEMServer %s' % er)
        sys.exit(1)

    # Create the listener and listener call back and start the listener
    #pylint: disable=global-statement
    global LISTENER
    ####stream=sys.stderr,
    logging.basicConfig(filename='pegasusindicationtest.log',
                        level=logging.INFO,
                        format='%(asctime)s %(levelname)s: %(message)s')
    # Create and start local listener
    LISTENER = WBEMListener(listener_host, http_port=http_listener_port,
                            https_port=https_listener_port)

    # Start connect and start listener.
    LISTENER.add_callback(consume_indication)
    LISTENER.start()

    listener_url = '%s://%s:%s' % ('http', 'localhost', http_listener_port)
    sub_mgr.add_listener_destinations(server_id, listener_url)

    # Create a dynamic alert indication filter and subscribe for it
    filter_ = sub_mgr.add_filter(
        server_id, TEST_CLASS_NAMESPACE,
        TEST_QUERY,
        query_language="DMTF:CQL")
    subscriptions = sub_mgr.add_subscriptions(server_id, filter_.path)

    # Request server to create indications by invoking method
    # This is pegasus specific
    class_name = CIMClassName(TEST_CLASS, namespace=TEST_CLASS_NAMESPACE)

    while repeat_loop > 0:
        repeat_loop += -1
        global RECEIVED_INDICATION_COUNT, INDICATION_START_TIME
        RECEIVED_INDICATION_COUNT = 0
        INDICATION_START_TIME = None
        if send_request_for_indications(conn, class_name,
                                        requested_indications):
            # Wait for indications to be received.
            success = wait_for_indications(requested_indications)
            if not success:
                insts = conn.EnumerateInstances('PG_ListenerDestinationQueue',
                                                namespace='root/PG_Internal')
                for inst in insts:
                    print('%s queueFullDropped %s, maxretry %s, InQueue %s' % \
                          (inst['ListenerDestinationName'],
                           inst['QueueFullDroppedIndications'],
                           inst['RetryAttemptsExceededIndications'],
                           inst['CurrentIndications']))
        if repeat_loop > 0:
            time.sleep(requested_indications/150)


    sub_mgr.remove_subscriptions(server_id,
                                 [inst.path for inst in subscriptions])
    sub_mgr.remove_filter(server_id, filter_.path)
    sub_mgr.remove_server(server_id)
    LISTENER.stop()

    # Test for all expected indications received.
    if RECEIVED_INDICATION_COUNT != requested_indications:
        print('Incorrect count of indications received expected=%s, received'
              '=%s' % (requested_indications, RECEIVED_INDICATION_COUNT))
        sys.exit(1)
    else:
        print('Success, %s indications' % requested_indications)
    print('Max time between indications %s' % MAX_TIME_BETWEEN_INDICATIONS)


def main():
    """Setup parameters for the test and call the test function
        This is a very simple interface with fixed cli arguments. If there
        are no arguments it defaults to a standard internal set of
        arguments
    """

    if len(sys.argv) < 7:
        print("Requires fixed set of arguments or defaults to internally\n "
              "defined arguments.\n"
              "Usage: %s <url> <username> <password> <indication-count>" \
              "Where: <url> server url, ex. http://localhost\n" \
              "       <port> http listener port, ex. 5000\n" \
              "       <username> username for authentication\n" \
              "       <password> password for authentication\n" \
              "       <indication-count> Number of indications to request.\n" \
              "       <repeat_loop>   Repeat the send test repeat_loop times.\n"
              "Ex: %s http://fred 5000 blah blah 1000 10 " \
              % (sys.argv[0], sys.argv[0]))
        server_url = 'http://localhost'
        username = 'blah'
        password = 'blah'
        http_listener_port = 5000
        requested_indications = 1000
        repeat_loop = 1

    else:
        server_url = sys.argv[1]
        http_listener_port = int(sys.argv[2])
        username = sys.argv[3]
        password = sys.argv[4]
        requested_indications = int(sys.argv[5])
        repeat_loop = int(sys.argv[6])

    listener_addr = urlparse(server_url).netloc

    print('url=%s listener=%s port=%s usr=%s pw=%s cnt=%s repeat=%s' % \
          (server_url, listener_addr, http_listener_port, \
           username, password, requested_indications, repeat_loop))

    #https_listener_port = http_listener_port + 1
    https_listener_port = None

    run_test(server_url, listener_addr, username, password,
             http_listener_port, https_listener_port,
             requested_indications, repeat_loop)

    return 0


if __name__ == '__main__':
    sys.exit(main())



#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

"""
*New in pywbem 0.9 as experimental and finalized in 0.10.*

The :class:`~pywbem.WBEMListener` class provides a thread-based WBEM listener
service that can receive CIM indications from multiple WBEM servers and that
calls registered callback functions to deliver the received indications.

Examples
--------

The following example creates and runs a listener::

    import sys
    import logging
    from socket import getfqdn
    from pywbem import WBEMListener

    def process_indication(indication, host):
        '''This function gets called when an indication is received.'''

        print("Received CIM indication from {host}: {ind!r}". \\
            format(host=host, ind=indication))

    def main():

        # Configure logging of the listener via the Python root logger
        logging.basicConfig(filename='listener.log', level=logging.WARNING,
            format='%(asctime)s - %(levelname)s - %(message)s)

        certkeyfile = 'listener.pem'

        listener = WBEMListener(host=getfqdn()
                                http_port=5988,
                                https_port=5989,
                                certfile=certkeyfile,
                                keyfile=certkeyfile)
        listener.add_callback(process_indication)

        try:
            listener.start()

            # process_indication() will be called for each received indication

            ... # wait for some condition to end listening

        finally:
            listener.stop()

Alternative code using the class as a context manager::

    with WBEMListener(...) as listener:
        listener.add_callback(process_indication)
        listener.start()

        # process_indication() will be called for each received indication

        ... # wait for some condition to end listening

    # listener.stop() has been called automatically

See the example in section :ref:`WBEMSubscriptionManager` for an example of
using a listener in combination with a subscription manager.

Another listener example is in the script ``examples/listen.py`` (when you
clone the GitHub pywbem/pywbem project). It is an interactive Python shell that
creates a listener and displays any indications it receives, in MOF format.


.. _`Logging in the listener`:

Logging in the listener
-----------------------

Each :class:`~pywbem.WBEMListener` object has its own separate Python logger
object with the name:

  `'pywbem.listener.{id}'`

where `{id}` is a string that is unique for each :class:`~pywbem.WBEMListener`
object within the Python process.

The :attr:`~pywbem.WBEMListener.logger` property of a
:class:`~pywbem.WBEMListener` object provides access to that Python logger
object, if needed.

The listener will log any indications it receives and any responses it sends
back to the indication sender, at the :attr:`py:logging.INFO` logging level.

In addition, it will log errors at the :attr:`py:logging.ERROR` logging level.

Since Python 2.7, the Python root logger will by default (i.e. when not being
configured) print log records of logging level :attr:`py:logging.WARNING` or
greater to `sys.stderr`. So the indication and response interactions will not
be printed by default, but any errors logged at the :attr:`py:logging.ERROR`
logging level will be printed by default.

Pywbem adds a null handler to the logger named `'pywbem'`, in order to prevent
the "No handlers could be found for logger ..." warning.
This follows best practices recommended in `Configuring logging for a library
<https://docs.python.org/2/howto/logging.html#configuring-logging-for-a-library>`_
and in several articles, for example in `this article
<http://pieces.openpolitics.com/2012/04/python-logging-best-practices/>`_.
Because this warning is no longer issued on Python 3.4 and higher, pywbem
adds a null handler only on Python 2.6 and 2.7.


.. _`WBEMListener class`:

WBEMListener class
------------------
"""

import sys
import errno
import re
import logging
import ssl
import threading
import six
from six.moves import BaseHTTPServer
from six.moves import socketserver
from six.moves import http_client

from . import cim_xml
from ._version import __version__
from .cim_obj import CIMInstance
from .cim_constants import CIM_ERR_NOT_SUPPORTED, CIM_ERR_INVALID_PARAMETER, \
    _statuscode2name
from .tupleparse import TupleParser
from .tupletree import xml_to_tupletree_sax
from .exceptions import CIMXMLParseError, XMLParseError, VersionError
from ._utils import _format

# CIM-XML protocol related versions implemented by the WBEM listener.
# These are returned in export message responses.
IMPLEMENTED_CIM_VERSION = '2.0'
IMPLEMENTED_DTD_VERSION = '2.4'
IMPLEMENTED_PROTOCOL_VERSION = '1.4'

# CIM-XML protocol related versions supported by the WBEM listener
# These are checked in export message requests.
SUPPORTED_DTD_VERSION_PATTERN = r'2\.\d+'
SUPPORTED_DTD_VERSION_STR = '2.x'
SUPPORTED_PROTOCOL_VERSION_PATTERN = r'1\.\d+'
SUPPORTED_PROTOCOL_VERSION_STR = '1.x'

# Pattern for findall() for header values that are a list of tokens with
# quality values (see RFC2616). The pattern does not verify conformance
# to the valid characters for tokens, but does its job in parsing tokens
# and q values.
TOKEN_QUALITY_FINDALL_PATTERN = re.compile(
    r'([^;, ]+)'
    r'(?:; *q=([01](?:\.[0-9]*)?))?'
    r'(?:, *)?')
TOKEN_CHARSET_FINDALL_PATTERN = re.compile(
    r'([^;, ]+)'
    r'(?:; *charset="?([^";, ]*)"?)?'
    r'(?:, *)?')

__all__ = ['WBEMListener', 'callback_interface']


class ThreadedHTTPServer(socketserver.ThreadingMixIn,
                         BaseHTTPServer.HTTPServer):
    """Defines an HTTPServer class for indication reception"""
    pass


class ListenerRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):
    """
    A request handler for the standard Python HTTP server, with a handler
    method for the HTTP POST method, that acts as a WBEM listener.
    """

    def invalid_method(self):
        """
        Handle invalid HTTP methods by sending HTTP status 405 "Method Not
        Allowed" back to the server. See DSP0200 for details on this.
        """
        self.send_http_error(405, headers=[('Allow', 'POST')])

    # pylint: disable=invalid-name
    def do_OPTIONS(self):
        """Invalid method for listener"""
        self.invalid_method()

    # pylint: disable=invalid-name
    def do_HEAD(self):
        """Invalid method for listener"""
        self.invalid_method()

    # pylint: disable=invalid-name
    def do_GET(self):
        """Invalid method for listener"""
        self.invalid_method()

    # pylint: disable=invalid-name
    def do_PUT(self):
        """Invalid method for listener"""
        self.invalid_method()

    # pylint: disable=invalid-name
    def do_PATCH(self):
        """Invalid method for listener"""
        self.invalid_method()

    # pylint: disable=invalid-name
    def do_DELETE(self):
        """Invalid method for listener"""
        self.invalid_method()

    # pylint: disable=invalid-name
    def do_TRACE(self):
        """Invalid method for listener"""
        self.invalid_method()

    # pylint: disable=invalid-name
    def do_CONNECT(self):
        """Invalid method for listener"""
        self.invalid_method()

    # pylint: disable=invalid-name
    def do_M_POST(self):
        """Invalid method for listener"""
        self.invalid_method()

    # pylint: disable=invalid-name
    def do_POST(self):
        """
        This method will be called for each POST request to one of the
        listener ports.

        It parses the CIM-XML export message and delivers the contained
        CIM indication to the stored listener object.
        """

        # Accept header check described in DSP0200
        accept = self.headers.get('Accept', 'text/xml')
        if accept not in ('text/xml', 'application/xml', '*/*'):
            self.send_http_error(
                406, 'header-mismatch',
                _format("Invalid Accept header value: {0} (need text/xml, "
                        "application/xml or */*)", accept))
            return

        # Accept-Charset header check described in DSP0200
        accept_charset = self.headers.get('Accept-Charset', 'UTF-8')
        tq_list = re.findall(TOKEN_QUALITY_FINDALL_PATTERN, accept_charset)
        found = False
        if tq_list is not None:
            for token, quality in tq_list:
                if token.lower() in ('utf-8', '*'):
                    found = True
                    break
        if not found:
            self.send_http_error(
                406, 'header-mismatch',
                _format("Invalid Accept-Charset header value: {0} "
                        "(need UTF-8 or *)", accept_charset))
            return

        # Accept-Encoding header check described in DSP0200
        accept_encoding = self.headers.get('Accept-Encoding', 'Identity')
        tq_list = re.findall(TOKEN_QUALITY_FINDALL_PATTERN, accept_encoding)
        identity_acceptable = False
        identity_found = False
        if tq_list is not None:
            for token, quality in tq_list:
                quality = 1 if quality == '' else float(quality)
                if token.lower() == 'identity':
                    identity_found = True
                    if quality > 0:
                        identity_acceptable = True
                    break
            if not identity_found:
                for token, quality in tq_list:
                    quality = 1 if quality == '' else float(quality)
                    if token == '*' and quality > 0:
                        identity_acceptable = True
                        break
        if not identity_acceptable:
            self.send_http_error(
                406, 'header-mismatch',
                _format("Invalid Accept-Encoding header value: {0} "
                        "(need Identity to be acceptable)", accept_encoding))
            return

        # Accept-Language header check described in DSP0200.
        # Ignored, because this WBEM listener does not support multiple
        # languages, and hence any language is allowed to be returned.

        # Accept-Range header check described in DSP0200
        accept_range = self.headers.get('Accept-Range', None)
        if accept_range is not None:
            self.send_http_error(
                406, 'header-mismatch',
                _format("Accept-Range header is not permitted {0}",
                        accept_range))
            return

        # Content-Type header check described in DSP0200
        content_type = self.headers.get('Content-Type', None)
        if content_type is None:
            self.send_http_error(
                406, 'header-mismatch',
                "Content-Type header is required")
            return
        tc_list = re.findall(TOKEN_CHARSET_FINDALL_PATTERN, content_type)
        found = False
        if tc_list is not None:
            for token, charset in tc_list:
                if token.lower() in ('text/xml', 'application/xml') and \
                   (charset == '' or charset.lower() == 'utf-8'):
                    found = True
                    break
        if not found:
            self.send_http_error(
                406, 'header-mismatch',
                _format("Invalid Content-Type header value: {0} "
                        "(need text/xml or application/xml with "
                        "charset=utf-8 or empty)",
                        content_type))
            return

        # Content-Encoding header check described in DSP0200
        content_encoding = self.headers.get('Content-Encoding', 'identity')
        if content_encoding.lower() != 'identity':
            self.send_http_error(
                406, 'header-mismatch',
                _format("Invalid Content-Encoding header value: {0}"
                        "(listener supports only identity)",
                        content_encoding))
            return

        # Content-Language header check described in DSP0200.
        # Ignored, because this WBEM listener does not support multiple
        # languages, and hence any language is allowed in the request.

        # The following headers are ignored. They are not allowed to be used
        # by servers, but listeners are not required to reject them:
        # Content-Range, Expires, If-Range, Range.

        # Start processing the request
        content_len = int(self.headers.get('Content-Length', 0))
        body = self.rfile.read(content_len)

        try:
            msgid, methodname, params = self.parse_export_request(body)
        except (CIMXMLParseError, XMLParseError) as exc:
            self.send_http_error(400, "request-not-well-formed", str(exc))
            return
        except VersionError as exc:
            if str(exc).startswith("DTD"):
                self.send_http_error(400, "unsupported-dtd-version",
                                     str(exc))
            elif str(exc).startswith("Protocol"):
                self.send_http_error(400, "unsupported-protocol-version",
                                     str(exc))
            else:
                self.send_http_error(400, "unsupported-version", str(exc))
            return

        if methodname == 'ExportIndication':

            if len(params) != 1 or 'NewIndication' not in params:
                self.send_error_response(
                    msgid, methodname, CIM_ERR_INVALID_PARAMETER,
                    _format("Expecting one parameter NewIndication, got {0!A}",
                            params.keys()))
                return

            indication_inst = params['NewIndication']

            if not isinstance(indication_inst, CIMInstance):
                self.send_error_response(
                    msgid, methodname, CIM_ERR_INVALID_PARAMETER,
                    _format("NewIndication parameter is not a CIM instance, "
                            "but {0!A}", indication_inst))
                return
            # server.listener created in WBEMListener.start function
            self.server.listener.deliver_indication(indication_inst,
                                                    self.client_address[0])

            self.send_success_response(msgid, methodname)

        else:
            self.send_error_response(
                msgid, methodname, CIM_ERR_NOT_SUPPORTED,
                _format("Unknown export method: {0!A}", methodname))

    def send_http_error(self, http_code, cim_error=None,
                        cim_error_details=None, headers=None):
        """
        Send an HTTP response back to the WBEM server that indicates
        an error at the HTTP level.
        """
        self.send_response(http_code, http_client.responses.get(http_code, ''))
        self.send_header("CIMExport", "MethodResponse")
        if cim_error is not None:
            self.send_header("CIMError", cim_error)
        if cim_error_details is not None:
            self.send_header("CIMErrorDetails", cim_error_details)
        if headers is not None:
            for header, value in headers:
                self.send_header(header, value)
        self.end_headers()
        self.log('%s: HTTP status %s; CIMError: %s, CIMErrorDetails: %s',
                 (self._get_log_prefix(), http_code, cim_error,
                  cim_error_details),
                 logging.WARNING)

    def send_error_response(self, msgid, methodname, status_code, status_desc,
                            error_insts=None):
        """Send a CIM-XML response message back to the WBEM server that
        indicates error."""

        resp_xml = cim_xml.CIM(
            cim_xml.MESSAGE(
                cim_xml.SIMPLEEXPRSP(
                    cim_xml.EXPMETHODRESPONSE(
                        methodname,
                        cim_xml.ERROR(
                            str(status_code),
                            status_desc,
                            error_insts),
                        ),  # noqa: E123
                    ),  # noqa: E123
                msgid, IMPLEMENTED_PROTOCOL_VERSION),
            IMPLEMENTED_CIM_VERSION, IMPLEMENTED_DTD_VERSION)

        resp_body = '<?xml version="1.0" encoding="utf-8" ?>\n' + \
                    resp_xml.toxml()

        if isinstance(resp_body, six.text_type):
            resp_body = resp_body.encode("utf-8")

        http_code = 200
        self.send_response(http_code, http_client.responses.get(http_code, ''))
        self.send_header("Content-Type", "text/html")
        self.send_header("Content-Length", str(len(resp_body)))
        self.send_header("CIMExport", "MethodResponse")
        self.end_headers()
        self.wfile.write(resp_body)
        self.log('%s: HTTP status %s; CIM error response: %s: %s',
                 (self._get_log_prefix(), http_code,
                  _statuscode2name(status_code), status_desc),
                 logging.WARNING)

    def send_success_response(self, msgid, methodname):
        """Send a CIM-XML response message back to the WBEM server that
        indicates success."""

        resp_xml = cim_xml.CIM(
            cim_xml.MESSAGE(
                cim_xml.SIMPLEEXPRSP(
                    cim_xml.EXPMETHODRESPONSE(
                        methodname),
                    ),  # noqa: E123
                msgid, IMPLEMENTED_PROTOCOL_VERSION),
            IMPLEMENTED_CIM_VERSION, IMPLEMENTED_DTD_VERSION)
        resp_body = '<?xml version="1.0" encoding="utf-8" ?>\n' + \
                    resp_xml.toxml()

        if isinstance(resp_body, six.text_type):
            resp_body = resp_body.encode("utf-8")

        http_code = 200
        self.send_response(http_code, http_client.responses.get(http_code, ''))
        self.send_header("Content-Type", "text/html")
        self.send_header("Content-Length", str(len(resp_body)))
        self.send_header("CIMExport", "MethodResponse")
        self.end_headers()
        self.wfile.write(resp_body)

    @staticmethod
    def parse_export_request(request_str):
        """Parse a CIM-XML export request message, and return
        a tuple(msgid, methodname, params).
        """

        # Parse the XML into a tuple tree (may raise CIMXMLParseError or
        # XMLParseError):

        tt_ = xml_to_tupletree_sax(request_str, "CIM-XML export request")
        tp = TupleParser()
        tup_tree = tp.parse_cim(tt_)

        # Check the tuple tree

        if tup_tree[0] != 'CIM':
            raise CIMXMLParseError(
                _format("Expecting CIM element, got {0}", tup_tree[0]))
        dtd_version = tup_tree[1]['DTDVERSION']
        if not re.match(SUPPORTED_DTD_VERSION_PATTERN, dtd_version):
            raise VersionError(
                _format("DTD version {0} not supported. Supported versions "
                        "are: {1!A}",
                        dtd_version, SUPPORTED_DTD_VERSION_STR))
        tup_tree = tup_tree[2]

        if tup_tree[0] != 'MESSAGE':
            raise CIMXMLParseError(
                _format("Expecting MESSAGE element, got {0}", tup_tree[0]))
        msgid = tup_tree[1]['ID']
        protocol_version = tup_tree[1]['PROTOCOLVERSION']
        if not re.match(SUPPORTED_PROTOCOL_VERSION_PATTERN, protocol_version):
            raise VersionError(
                _format("Protocol version {0} not supported. "
                        "Supported versions are: {1!A}",
                        protocol_version, SUPPORTED_PROTOCOL_VERSION_STR))
        tup_tree = tup_tree[2]

        if tup_tree[0] != 'SIMPLEEXPREQ':
            raise CIMXMLParseError(
                _format("Expecting SIMPLEEXPREQ element, got {0}",
                        tup_tree[0]))
        tup_tree = tup_tree[2]

        if tup_tree[0] != 'EXPMETHODCALL':
            raise CIMXMLParseError(
                _format("Expecting EXPMETHODCALL element, got {0}",
                        tup_tree[0]))

        methodname = tup_tree[1]['NAME']
        params = {}
        for name, obj in tup_tree[2]:
            params[name] = obj

        return (msgid, methodname, params)

    def log(self, format_, args, level=logging.INFO):
        """
        This function is called for anything that needs to get logged.
        It logs to the logger of this listener.

        It is not defined in the standard handler class; our version
        has an additional `level` argument that allows to control the
        logging level in the standard Python logging support.

        Another difference is that the variable arguments are passed
        in as a tuple.
        """
        self.server.listener.logger.log(level, format_, *args)

    # pylint: disable=redefined-builtin
    def log_message(self, format, *args):
        """
        In the standard handler class, this function is called for anything
        that needs to get logged (e.g. from :meth:`log_request`).

        We override it in order to use our own log function.
        """
        self.log(format, args, logging.INFO)

    def log_request(self, code='-', size='-'):
        """
        This function is called during :meth:`send_response`.

        We override it to get a little more information logged in a somewhat
        better format.
        """
        self.log('%s: HTTP status %s',
                 (self._get_log_prefix(), code),
                 logging.INFO)

    def _get_log_prefix(self):
        """Return the prefix components for a log entry"""
        return _format("{0} {1} from {2}",
                       self.request_version, self.command,
                       self.client_address[0])

    def version_string(self):
        """
        Overrides the inherited method to add the pywbem listener version.
        """
        return _format("pywbem-listener/{0} {1} {2} ",
                       __version__, self.server_version, self.sys_version)


# pylint: disable=too-many-instance-attributes
class WBEMListener(object):
    """
    *New in pywbem 0.9 as experimental and finalized in 0.10.*

    A WBEM listener.

    The listener supports starting and stopping threads that listen for
    CIM-XML ExportIndication messages using HTTP and/or HTTPS, and that pass
    any received indications on to registered callback functions.

    The listener must be stopped in order to free the TCP/IP port it listens
    on. Using this class as a context manager ensures that the listener is
    stopped when leaving the context manager scope.
    """

    def __init__(self, host, http_port=None, https_port=None,
                 certfile=None, keyfile=None):
        """
        Parameters:

          host (:term:`string`):
            IP address or host name at which this listener can be reached.

          http_port (:term:`string` or :term:`integer`):
            HTTP port at which this listener can be reached. Note that at
            least one port (HTTP or HTTPS) must be set.

            `None` means not to set up a port for HTTP.

          https_port (:term:`string` or :term:`integer`):
            HTTPS port at which this listener can be reached.

            `None` means not to set up a port for HTTPS.

          certfile (:term:`string`):
            File path of certificate file to be used as server certificate
            during SSL/TLS handshake when creating the secure HTTPS connection.

            It is valid for the certificate file to contain a private key; the
            server certificate sent during SSL/TLS handshake is sent without
            the private key.

            `None` means not to use a server certificate file. Setting up a port
            for HTTPS requires specifying a certificate file.

          keyfile (:term:`string`):
            File path of private key file to be used by the server during
            SSL/TLS handshake when creating the secure HTTPS connection.

            It is valid to specify a certificate file that contains a private
            key.

            `None` means not to use a private key file. Setting up a port
            for HTTPS requires specifying a private key file.
        """

        self._host = host

        if isinstance(http_port, six.integer_types):
            self._http_port = int(http_port)  # Convert Python 2 long to int
        elif isinstance(http_port, six.string_types):
            self._http_port = int(http_port)
        elif http_port is None:
            self._http_port = http_port
        else:
            raise TypeError(
                _format("Invalid type for http_port: {0}", type(http_port)))

        if isinstance(https_port, six.integer_types):
            self._https_port = int(https_port)  # Convert Python 2 long to int
        elif isinstance(https_port, six.string_types):
            self._https_port = int(https_port)
        elif https_port is None:
            self._https_port = https_port
        else:
            raise TypeError(
                _format("Invalid type for https_port: {0}", type(https_port)))

        if self._https_port is not None:
            if certfile is None:
                raise ValueError("https_port requires certfile")
            self._certfile = certfile
            if keyfile is None:
                raise ValueError("https_port requires keyfile")
            self._keyfile = keyfile
        else:
            self._certfile = None
            self._keyfile = None

        if self._http_port is None and self._https_port is None:
            ValueError("Listener requires at least one active port")

        self._http_server = None  # ThreadedHTTPServer for HTTP
        self._http_thread = None  # Thread for HTTP
        self._https_server = None  # ThreadedHTTPServer for HTTPS
        self._https_thread = None  # Thread for HTTPS

        self._logger = logging.getLogger(
            _format("pywbem.listener.{0}", id(self)))

        self._callbacks = []  # Registered callback functions

    def __str__(self):
        """
        Return a representation of the :class:`~pywbem.WBEMListener` object
        with a subset of its attributes.
        """
        return _format(
            "WBEMListener("
            "_host={s._host!A}, "
            "_http_port={s._http_port!A}, "
            "_https_port={s._https_port!A}, "
            "...)",
            s=self)

    def __repr__(self):
        """
        Return a representation of the :class:`~pywbem.WBEMListener` object
        with all attributes, that is suitable for debugging.
        """
        return _format(
            "WBEMListener("
            "_host={s._host!A}, "
            "_http_port={s._http_port!A}, "
            "_https_port={s._https_port!A}, "
            "_certfile={s._certfile!A}, "
            "_keyfile={s._keyfile!A}, "
            "_logger={s._logger!A}, "
            "_callbacks={s._callbacks!A})",
            s=self)

    def __enter__(self):
        """
        *New in pywbem 0.12.*

        Enter method when the class is used as a context manager.

        Returns the listener object.
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        *New in pywbem 0.12.*

        Exit method when the class is used as a context manager.

        Stops the listener by calling :meth:`~pywbem.WBEMListener.stop`.
        """
        self.stop()
        return False  # re-raise any exceptions

    @property
    def host(self):
        """
        :term:`string`: IP address or host name at which this listener can be
        reached.
        """
        return self._host

    @property
    def http_port(self):
        """
        :term:`integer`: HTTP port at which this listener can be reached.

        `None` means there is no port set up for HTTP.
        """
        return self._http_port

    @property
    def https_port(self):
        """
        :term:`integer`: HTTPS port at which this listener can be reached.

        `None` means there is no port set up for HTTPS.
        """
        return self._https_port

    @property
    def http_started(self):
        """
        :class:`py:bool`: Boolean indicating whether the listener is started
        for the HTTP port.

        If no port is set up for HTTP, `False` is returned.

        *New in pywbem 0.12.*
        """
        return self._http_server is not None

    @property
    def https_started(self):
        """
        :class:`py:bool`: Boolean indicating whether the listener is started
        for the HTTPS port.

        If no port is set up for HTTPS, `False` is returned.

        *New in pywbem 0.12.*
        """
        return self._https_server is not None

    @property
    def certfile(self):
        """
        :term:`string`: File path of the certificate file used as server
        certificate during SSL/TLS handshake when creating the secure HTTPS
        connection.

        `None` means there is no certificate file being used (that is, no port
        is set up for HTTPS).
        """
        return self._certfile

    @property
    def keyfile(self):
        """
        :term:`string`: File path of the private key file used by the server
        during SSL/TLS handshake when creating the secure HTTPS connection.

        `None` means there is no certificate file being used (that is, no port
        is set up for HTTPS).
        """
        return self._keyfile

    @property
    def logger(self):
        """
        :class:`py:logging.Logger`: Logger object for this listener.

        Each listener object has its own separate logger object with the name:

          `'pywbem.listener.{id}'`

        where `{id}` is a unique string for each listener object.

        Users of the listener should not look up the logger object by name, but
        should use this property to get to it.
        """
        return self._logger

    def start(self):
        """
        Start the WBEM listener threads, if they are not yet running.

        A thread serving CIM-XML over HTTP is started if an HTTP port was
        specified for the listener.
        A thread serving CIM-XML over HTTPS is started if an HTTPS
        port was specified for the listener.

        These server threads will handle the ExportIndication export message
        described in :term:`DSP0200` and they will invoke the registered
        callback functions for any received CIM indications.

        The listener must be stopped again in order to free the TCP/IP port it
        listens on. The listener can be stopped explicitly using the
        :meth:`~pywbem.WBEMListener.stop` method. The listener will be
        automatically stopped when the main thread terminates (i.e. when the
        Python process terminates), or when :class:`~pywbem.WBEMListener`
        is used as a context manager when leaving its scope.

        Raises:

          :exc:`~py:exceptions.OSError`:
            with :attr:`~OSError.errno` =
            :data:`py:errno.EADDRINUSE` when the WBEM listener port is already
            in use.
        """

        if self._http_port:
            if not self._http_server:
                try:
                    server = ThreadedHTTPServer((self._host, self._http_port),
                                                ListenerRequestHandler)
                except Exception as exc:
                    # Linux+py2: socket.error; Linux+py3: OSError;
                    # Windows does not raise any exception.
                    if getattr(exc, 'errno', None) == errno.EADDRINUSE:
                        # Reraise with improved error message
                        msg = _format("WBEM listener port {0} already in use",
                                      self._http_port)
                        exc_type = OSError
                        six.reraise(exc_type, exc_type(errno.EADDRINUSE, msg),
                                    sys.exc_info()[2])
                    raise

                # pylint: disable=attribute-defined-outside-init
                server.listener = self
                thread = threading.Thread(target=server.serve_forever)
                thread.daemon = True  # Exit server thread upon main thread exit
                self._http_server = server
                self._http_thread = thread
                thread.start()
        else:
            # Just in case someone changed self._http_port after init...
            self._http_server = None
            self._http_thread = None

        if self._https_port:
            if not self._https_server:
                try:
                    server = ThreadedHTTPServer((self._host, self._https_port),
                                                ListenerRequestHandler)
                except Exception as exc:
                    # Linux+py2: socket.error; Linux+py3: OSError;
                    # Windows does not raise any exception.
                    if getattr(exc, 'errno', None) == errno.EADDRINUSE:
                        # Reraise with improved error message
                        msg = _format("WBEM listener port {0} already in use",
                                      self._http_port)
                        exc_type = OSError
                        six.reraise(exc_type, exc_type(errno.EADDRINUSE, msg),
                                    sys.exc_info()[2])
                    raise

                # pylint: disable=attribute-defined-outside-init
                server.listener = self
                server.socket = ssl.wrap_socket(server.socket,
                                                certfile=self._certfile,
                                                keyfile=self._keyfile,
                                                server_side=True)
                thread = threading.Thread(target=server.serve_forever)
                thread.daemon = True  # Exit server thread upon main thread exit
                self._https_server = server
                self._https_thread = thread
                thread.start()
        else:
            # Just in case someone changed self._https_port after init...
            self._https_server = None
            self._https_thread = None

    def stop(self):
        """
        Stop the WBEM listener threads, if they are running.
        """

        # Stopping the server will cause its `serve_forever()` method
        # to return, which will cause the server thread to terminate.
        # TODO: Describe how the processing threads terminate.

        if self._http_server:
            self._http_server.shutdown()
            self._http_server.server_close()
            self._http_server = None
            self._http_thread = None

        if self._https_server:
            self._https_server.shutdown()
            self._https_server.server_close()
            self._https_server = None
            self._https_thread = None

    def deliver_indication(self, indication, host):
        """
        This function is called by the listener threads for each received
        indication. It is not supposed to be called by the user.

        It delivers the indication to all callback functions that have been
        added to the listener.

        If a callback function raises any exception this is logged as an error
        using the listener logger and the next registered callback function is
        called.

        Parameters:

          indication (:class:`~pywbem.CIMInstance`):
            Representation of the CIM indication to be delivered.

          host (:term:`string`):
            Host name or IP address of WBEM server sending the indication.
        """
        for callback in self._callbacks:
            try:
                callback(indication, host)
            except Exception as exc:  # pylint: disable=broad-except
                self.logger.log(logging.ERROR, "Indication delivery callback "
                                "function raised %s: %s",
                                exc.__class__.__name__, exc)

    def add_callback(self, callback):
        """
        Add a callback function to the listener.

        The callback function will be called for each indication this listener
        receives from any WBEM server.

        If the callback function is already known to the listener, it will not
        be added.

        Parameters:

          callback (:func:`~pywbem.callback_interface`):
            Callable that is being called for each CIM indication that is
            received while the listener threads are running.
        """
        if callback not in self._callbacks:
            self._callbacks.append(callback)


def callback_interface(indication, host):
    # pylint: disable=unused-argument
    """
    *New in pywbem 0.9 as experimental and finalized in 0.10.*

    Interface of a callback function that is provided by the user of the API
    and that will be called by the listener for each received CIM indication.

    Parameters:

      indication (:class:`~pywbem.CIMInstance`):
        Representation of the CIM indication that has been received.
        Its `path` attribute is `None`.

      host (:term:`string`):
        Host name or IP address of WBEM server sending the indication.

    Raises:

        Exception: If a callback function raises any exception this is logged
          as an error using the listener logger and the next registered
          callback function is called.
    """
    raise NotImplementedError



#!/usr/bin/env python

"""
    Python script to create and send indications to a listener. Allows sending
    either http or https requests.  Note that the verify is forced to
    false for this version so that certificates or keys are not required.
"""

from __future__ import print_function, absolute_import
import sys
from time import time
import datetime
import argparse as _argparse
import re
from random import randint
from pywbem._cliutils import SmartFormatter as _SmartFormatter
import requests
# The following disables the urllib3 InsecureRequestWarning that gets
# generated for every send when server cert verification is disabled.
from requests.packages.urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
import xml.etree.ElementTree as ET
import json

class ElapsedTimer(object):
    """
        Set up elapsed time timer. Calculates time between initiation
        and access.
    """
    def __init__(self):
        """ Initiate the object with current time"""
        self.start_time = datetime.datetime.now()

    def reset(self):
        """ Reset the start time for the timer"""
        self.start_time = datetime.datetime.now()

    def elapsed_ms(self):
        """ Get the elapsed time in milliseconds. returns floating
            point representation of elapsed time in seconds.
        """
        dt = datetime.datetime.now() - self.start_time
        return ((dt.days * 24 * 3600) + dt.seconds) * 1000  \
                + dt.microseconds / 1000.0

    def elapsed_sec(self):
        """ get the elapsed time in seconds. Returns floating
            point representation of time in seconds
        """
        return self.elapsed_ms() / 1000


def create_indication_data(msg_id, sequence_number, source_id, delta_time, \
                           protocol_ver):
    '''
    Create a singled indication XML from the template and the included
    sequence_number, delta_time, and protocol_ver

    Returns the completed indication XML
    '''

    data_template = """<?xml version="1.0" encoding="utf-8" ?>
    <CIM CIMVERSION="2.0" DTDVERSION="2.4">
      <MESSAGE ID="%(msg_id)s" PROTOCOLVERSION="%(protocol_ver)s">
        <SIMPLEEXPREQ>
          <EXPMETHODCALL NAME="ExportIndication">
            <EXPPARAMVALUE NAME="NewIndication">
              <INSTANCE CLASSNAME="CIM_AlertIndication">
                <PROPERTY NAME="Severity" TYPE="string">
                  <VALUE>high</VALUE>
                </PROPERTY>
                <PROPERTY NAME="SequenceNumber" TYPE="string">
                  <VALUE>%(sequence_number)s</VALUE>
                </PROPERTY>
                <PROPERTY NAME="SourceId" TYPE="string">
                  <VALUE>%(source_id)s</VALUE>
                </PROPERTY>
                <PROPERTY NAME="DELTA_TIME" TYPE="string">
                  <VALUE>%(delta_time)s</VALUE>
                </PROPERTY>
              </INSTANCE>
            </EXPPARAMVALUE>
          </EXPMETHODCALL>
        </SIMPLEEXPREQ>
      </MESSAGE>
    </CIM>"""

    data = {'sequence_number' : sequence_number, 'source_id' : source_id,
            'delta_time' : delta_time, 'protocol_ver' : protocol_ver,
            'msg_id' : msg_id}
    return data_template%data


def send_indication(url, headers, payload, verbose, verify=None, keyfile=None,
                    timeout=4):
    '''
    Send indication using requests module.

    Parameters:

    url(:term:`string):
      listener url including scheme, host name, port

    headers:
      All headers for the request

    pyload:
       XML payload containing the indication

    verbose:
      Flag for verbose responses

    verify:
      Either False or file of cert for verification of host. Note that
      this combination is unique to the requests module in place of using
      a no_verification flag.

    keyfile:
      None if there is no client verification or either a single
      file containing the cert amd private key file or a pair of files
      containing both (key, cert)
    
    Returns:

      True if response code = 200. Otherwise False
    '''

    try:
        response = requests.post(url, headers=headers, data=payload,
                                 timeout=timeout,
                                 verify=verify)

    except Exception as ex:
        print('Exception %s' % ex)
        return False

    # validate response status code and xml
    if verbose or response.status_code != 200:
        print('\nResponse code=%s headers=%s data=%s' % (response.status_code,
                                                         response.headers,
                                                         response.text))
    root = ET.fromstring(response.text)
    if (root.tag != 'CIM') or (root.attrib['CIMVERSION'] != '2.0') \
                           or (root.attrib['DTDVERSION'] != '2.4'):

        print('Invalid XML\nResponse code=%s headers=%s data=%s' % \
            (response.status_code, response.headers, response.text))
        return False
    for child in root:
        if child.tag != 'MESSAGE' or child.attrib['PROTOCOLVERSION'] != '1.4':
            print('Invalid child\nResponse code=%s headers=%s data=%s' % \
                  (response.status_code, response.headers, response.text))
            return False

    return True if(response.status_code == 200) else False

def get_args():
    """
        Get the command line arguments. use --help to get info on cmd line
        arguments
    """

    prog = "sendIndications"
    usage = '%(prog)s [options] listener-url'
    desc = 'Send indications to a listener. Verify set to False'
    epilog = """
Examples:
  %s https://127.0.0.1 -p 5001
  %s http://localhost:5000
  %s http://[2001:db8::1234-eth0] -(http port 5988 ipv6, zone id eth0)
""" % (prog, prog, prog)

    argparser = _argparse.ArgumentParser(
        prog=prog, usage=usage, description=desc, epilog=epilog,
        add_help=False, formatter_class=_SmartFormatter)

    pos_arggroup = argparser.add_argument_group(
        'Positional arguments')

    pos_arggroup.add_argument(
        'url', metavar='url', nargs='?', default='http://127.0.0.1',
        help='Optional url of listener '
             'If supplied, must be schema+address[:port]. Schema determines ' \
             ' whether http or https are used. Port may be defined in url ' \
             'or with -p option. If not included default port = 5000')

    general_arggroup = argparser.add_argument_group('General options')
    general_arggroup.add_argument(
        '-v', '--verbose', dest='verbose',
        action='store_true', default=False,
        help='Print more messages while processing')

    general_arggroup.add_argument(
        '--listenerPort', '-p', default=None,
        metavar="port",
        type=int,
        help='Integer argument defines listener port.')
    general_arggroup.add_argument(
        '--deliver', '-d', default=1,
        metavar='integer',
        type=int,
        help='Integer argument defines number of indications to send.')

    general_arggroup.add_argument(
        '-h', '--help', action='help',
        help='Show this help message and exit')

    security_arggroup = argparser.add_argument_group(
        'Connection security related options',
        'Specify user name and password or certificates and keys')
    security_arggroup.add_argument(
        '--certfile', '-c', dest='cert_file', metavar='certfile',
        help='R|Client certificate file for authenticating with the\n' \
             'WBEM listener. If option specified the client attempts\n' \
             'to execute server authentication.\n'
             'Default: None so that there is no verification of server cert.')
    security_arggroup.add_argument(
        '--keyfile', '-k', dest='key_file', metavar='keyfile',
        help='R|Client private key file for authenticating with the\n' \
             'WBEM listener. Specify as a single file (containing private ' \
             ' key and certificate or a tuple of key and certificate files')

    args = argparser.parse_args()

    if re.match('^http', args.url) is None:
        print('ERROR: url must include schema. received %s' % args.url)
        sys.exit(1)

    if args.verbose:
        print('args: %s' % args)

    return args, argparser

def main():
    """ Get arguments from get_args and create/send the number
        of indications defined.  Each indication is created from a
        template.
    """
    opts, argparser = get_args()

    start_time = time()

    url = opts.url
    if re.search(r":([0-9]+)$", opts.url):
        if opts.listenerPort is not None:
            argparser.error('Simultaneous url with port and -p port option '
                            'invalid')
    else:
        if opts.listenerPort is None:
            url = '%s:%s' % (opts.url, 5000)
        else:
            url = '%s:%s' % (opts.url, opts.listenerPort)

    if opts.verbose:
        print('url=%s' % url)

    cim_protocol_version = '1.4'

    # requests module combines the verification flag and certfile attribute
    # If verify=False, there is no verification of the server cert. If
    # verify=<file_name or dir name> it is the directory of the cert to
    # use for verification.
    verification = False if opts.cert_file is None else opts.cert_file

    headers = {'content-type' : 'application/xml; charset=utf-8',
               'CIMExport' : 'MethodRequest',
               'CIMExportMethod' : 'ExportIndication',
               'Accept-Encoding' : 'Identity',
               'CIMProtocolVersion' : cim_protocol_version}
    # includes accept-encoding because of requests issue.  He supplies it if
    # we don't TODO try None

    delta_time = time() - start_time
    rand_base = randint(1, 1000)
    timer = ElapsedTimer()
    source_id = 'send_indications.py'
    for i in range(opts.deliver):

        msg_id = '%s' % (i + rand_base)
        payload = create_indication_data(msg_id, i, source_id, delta_time,
                                         cim_protocol_version)

        if opts.verbose:
            print('headers=%s\n\npayload=%s' % (headers, payload))

        success = send_indication(url, headers, payload, opts.verbose,
                                  verify=verification)

        if success:
            if opts.verbose:
                print('sent # %s' % i)
            else:
                if i % 100 == 0:
                    sys.stdout.write('.')
                    sys.stdout.flush()
        else:
            print('Error return from send. Terminating.')
            return
    endtime = timer.elapsed_sec()
    print('Sent %s in %s sec or %.2f ind/sec' % (opts.deliver, endtime,
                                                 (opts.deliver/endtime)))

if __name__ == '__main__':
    sys.exit(main())



"""
This wbemcli scriptlet creates a WBEMServer for the connection
This is not a separate script but a scriplet for wbemcli.
It implements the creation of a WBEMServer based on the CONN, the
creation of the org_valuemap for profiles and a print function to
display basic profile information.
"""

from pywbem import WBEMServer, ValueMapping


def print_profile_info(org_vm, profile_instance):
    """
    Print information on a profile defined by profile_instance.

    Parameters:

      org_vm: The value mapping for CIMRegisterdProfile and
          RegisteredOrganization so that the value and not value mapping
          is displayed.

      profile_instance: instance of a profile to be printed
    """
    org = org_vm.tovalues(profile_instance['RegisteredOrganization'])
    name = profile_instance['RegisteredName']
    vers = profile_instance['RegisteredVersion']
    print("  %s %s Profile %s" % (org, name, vers))

# Create the server and save in global SERVER
SERVER = WBEMServer(CONN)

# create the CIMRegisterd Profile ValueMapping for the
# defined server. This can be used to
org_vm = ValueMapping.for_property(SERVER, SERVER.interop_ns,
                                   'CIM_RegisteredProfile',
                                   'RegisteredOrganization')

print("Brand:\n  %s" % SERVER.brand)
print("Version:\n  %s" % SERVER.version)
print("Interop namespace:\n  %s" % SERVER.interop_ns)

for inst in SERVER.profiles:
    print_profile_info(org_vm, inst)



#
# (C) Copyright 2004,2006 Hewlett-Packard Development Company, L.P.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# Author: Tim Potter <tpot@hp.com>
# Author: Martin Pool <mbp@hp.com>
#

"""
Pywbem is a WBEM client, written in pure Python.
It supports Python 2 and Python 3.
"""

# There are submodules, but clients shouldn't need to know about them.
# Importing just this module is enough.
# These are explicitly safe for 'import *'

from __future__ import absolute_import

import sys
import logging

# The config namespace is imported as a sub-namespace to make the config
# variables accessible directly via their defining namespace. Importing
# them into the pywbem namespace would duplicate their names and thus
# would cause changes to the config variables not to be visible in their
# original namespace.
from . import config  # noqa: F401

from ._utils import *  # noqa: F403,F401
from .cim_types import *  # noqa: F403,F401
from .cim_constants import *  # noqa: F403,F401
from .cim_operations import *  # noqa: F403,F401
from ._nocasedict import *  # noqa: F403,F401
from .cim_obj import *  # noqa: F403,F401
from .tupleparse import *  # noqa: F403,F401
from .cim_http import *  # noqa: F403,F401
from .exceptions import *  # noqa: F403,F401
from .mof_compiler import *  # noqa: F403,F401
from ._valuemapping import *  # noqa: F403,F401
from ._server import *  # noqa: F403,F401
from ._subscription_manager import *  # noqa: F403,F401
from ._listener import *  # noqa: F403,F401
from ._recorder import *  # noqa: F403,F401
from ._statistics import *  # noqa: F403,F401
from ._logging import *  # noqa: F403,F401
from ._warnings import *  # noqa: F403,F401

from ._version import __version__  # noqa: F401

_python_m = sys.version_info[0]  # pylint: disable=invalid-name
_python_n = sys.version_info[1]  # pylint: disable=invalid-name
if _python_m == 2 and _python_n < 6:
    raise RuntimeError('On Python 2, pywbem requires Python 2.6 or higher')
elif _python_m == 3 and _python_n < 4:
    raise RuntimeError('On Python 3, pywbem requires Python 3.4 or higher')

# On Python 2, add a NullHandler to suppress the warning "No handlers could be
# found for logger ...".
if _python_m == 2:
    try:  # Python 2.7+ includes logging.NullHandler
        from logging import NullHandler
    except ImportError:
        class NullHandler(logging.Handler):
            """Implement logging NullHandler for Python 2.6"""
            def emit(self, record):
                pass
    logging.getLogger('pywbem').addHandler(NullHandler())



#!/usr/bin/env python

''' Test listener. When started this code creates a WBEMListener on the
    address/ports defined by the input parameters.  This listener
    displays any indications received as log entries.
'''

import sys as _sys
import os as _os
import code as _code
import errno as _errno
import logging
import threading
# Conditional support of readline module
try:
    import readline as _readline
    _HAVE_READLINE = True
except ImportError as arg:
    _HAVE_READLINE = False
import six

import pywbem

# For creating a self-signed certificate file with private key inside, issue:
#    openssl req -new -x509 -keyout server.pem -out server.pem -days 365 -nodes

# Dictionary to count indications received by host
RECEIVED_INDICATION_DICT = {}
COUNTER_LOCK = threading.Lock()

#pylint: disable=invalid-name
listener = None

def _process_indication(indication, host):
    '''This function gets called when an indication is received.'''

    global RECEIVED_INDICATION_DICT
    COUNTER_LOCK.acquire()
    if host in RECEIVED_INDICATION_DICT:
        RECEIVED_INDICATION_DICT[host] += 1
    else:
        RECEIVED_INDICATION_DICT[host] = 1
    COUNTER_LOCK.release()

    listener.logger.info("Consumed CIM indication #%s: host=%s\n%s",
                         RECEIVED_INDICATION_DICT, host, indication.tomof())

def _get_argv(index, default=None):
    ''' get the argv input argument defined by index. Return the default
        attribute if that argument does not exist
    '''
    return _sys.argv[index] if len(_sys.argv) > index else default


def status(reset=None):
    '''
        Show status of indications received. If optional reset attribute
        is True, reset the counter.
    '''
    global RECEIVED_INDICATION_DICT
    for host, count in six.iteritems(RECEIVED_INDICATION_DICT):
        print('Host %s Received %s indications' % (host, count))
        
    if reset:
        for host in RECEIVED_INDICATION_DICT:
            RECEIVED_INDICATION_DICT[host] = 0
            print('Host %s Reset: Received %s indications' %
                  (host, RECEIVED_INDICATION_DICT[host]))
        print('counts reset to 0')


def _main():
    ''' Main function gets input parameters, sets up listener and starts
        the listener
    '''
    global listener

    if len(_sys.argv) < 2 or _sys.argv[1] == '--help':
        print("Usage: %s host http_port [https_port certfile keyfile]" % \
              _sys.argv[0])
        _sys.exit(2)

    logging.basicConfig(stream=_sys.stderr, level=logging.INFO,
                        format='%(asctime)s %(levelname)s: %(message)s')

    host = _get_argv(1)
    http_port = _get_argv(2)
    https_port = _get_argv(3)
    certfile = _get_argv(4)
    keyfile = _get_argv(5)

    listener = pywbem.WBEMListener(host=host,
                                   http_port=http_port,
                                   https_port=https_port,
                                   certfile=certfile,
                                   keyfile=keyfile)

    listener.add_callback(_process_indication)
    listener.start()


    banner = """
WBEM listener started on host %s (HTTP port: %s, HTTPS port: %s).
This Python console displays any indications received by this listener as
logger outputs (Level=INFO) to the console by default.

listener: WBEMListener instance that has been started.

help(listener): Display help for using listener instance.

status() : Display count of indications received; status(reset=True) to reset
           counters

Modify logger characteristics through listener.logger:
     (ex. listener.logger.setLevel(logging.ERROR))

Ctrl-d to exit
""" % (host, http_port, https_port)

    # Determine file path of history file
    home_dir = '.'
    if 'HOME' in _os.environ:
        home_dir = _os.environ['HOME'] # Linux
    elif 'HOMEPATH' in _os.environ:
        home_dir = _os.environ['HOMEPATH'] # Windows
    histfile = '%s/.listen_history' % home_dir

    # Read previous command line history
    if _HAVE_READLINE:
        NotFoundError = getattr(__builtins__, 'FileNotFoundError', IOError)
        try:
            _readline.read_history_file(histfile)
        except NotFoundError as exc:
            if exc.errno != _errno.ENOENT:
                raise

    # Interact
    i = _code.InteractiveConsole(globals())
    i.interact(banner)

    # Save command line history
    if _HAVE_READLINE:
        _readline.write_history_file(histfile)

    return 0

if __name__ == '__main__':
    _sys.exit(_main())



"""
Assertion functions for pywbem end2end tests.
"""

from __future__ import absolute_import

import six

from pywbem import CIMInstance, CIMInstanceName
from pywbem._utils import _format

from .utils import path_equal, path_in, instance_of, ServerObjectCache

OBJECT_CACHE = ServerObjectCache()


def assert_number_of_instances_equal(
        conn, inst_list, inst_list_msg, exp_number):
    """
    Assert that the number of objects in a list of instances or instance paths
    is equal to an expected number.

    Parameters:

        conn (WBEMConnection with 'server_definition' attribute)

        exp_number (integer): Expected number of instances or instance paths.

        inst_list (list of CIMInstanceName or CIMInstance): List of instances
          or instance paths to be tested.

        inst_list_msg (string): Short definition what inst_list is.
    """
    if len(inst_list) != exp_number:
        raise AssertionError(
            _format("Server {0} at {1}: List of instances ({2}) does not have "
                    "the expected exact size of {3} but is {4}",
                    conn.server_definition.nickname, conn.url,
                    inst_list_msg, exp_number, len(inst_list)))


def assert_number_of_instances_minimum(
        conn, inst_list, inst_list_msg, exp_number):
    """
    Assert that the number of objects in a list of instances or instance paths
    is equal to or larger than an expected number.

    Parameters:

        conn (WBEMConnection with 'server_definition' attribute)

        exp_number (integer): Expected number of instances or instance paths.

        inst_list (list of CIMInstanceName or CIMInstance): List of instances
          or instance paths to be tested.

        inst_list_msg (string): Short definition what inst_list is.
    """
    if len(inst_list) < exp_number:
        raise AssertionError(
            _format("Server {0} at {1}: List of instances ({2}) does not have "
                    "the expected minimum size of {3} but is {4}",
                    conn.server_definition.nickname, conn.url,
                    inst_list_msg, exp_number, len(inst_list)))


def assert_instance_of(conn, path_list, classname):
    """
    Assert that a set of CIM instances (identified by its instance paths)
    is of a particular CIM class, i.e. that the creation class of the CIM
    instances or one of its superclasses has the specified class name.

    Because there are WBEM servers without support for class operations,
    this is implemented without relying on class operations. The function
    performs an EnumerateInstanceNames operation on the desired class in
    the namespace of the instance in question, and verifies that the
    instance in question is in the result.

    Parameters:

        conn (WBEMConnection with 'server_definition' attribute)

        path_list (CIMInstanceName or CIMInstance or tuple/list
          thereof): The CIM instance paths. The instance paths must have a
          namespace and classname set. The namespace must be the same
          across all instance paths. The host portion of the instance
          paths is treated specially when comparing them, as described
          in path_equal().

        classname (string): The CIM class name.
    """

    # Parameters are checked and transformed in instance_of()

    # TODO 2018-12 AM: Improve performance by avoiding EI on each path
    if not isinstance(path_list, (tuple, list)):
        path_list = [path_list]
    for path in path_list:
        if not instance_of(conn, path, classname):
            raise AssertionError(
                _format("Server {0} at {1}: Instance at {2!A} is not of "
                        "class {3!A}",
                        conn.server_definition.nickname, conn.url,
                        path.to_wbem_uri(), classname))


def assert_instance_consistency(conn, instance, path):
    """
    Assert that an instance and an instance path are consistent:

    - They must have the same classname.
    - The instance must have the key properties matching the keybindings
      of the instance path.

    The 'path' attribute of the instance is ignored for this purpose.
    Of course, this function can be used to verify consistency of
    an instance and its 'path' attribute by passing it as the 'path'
    parameter.

    Parameters:

        conn (WBEMConnection with 'server_definition' attribute)

        instance (CIMInstance): The CIM instance to be verified.

        path (CIMInstanceName): The CIM instance path to be verified.
    """

    # Check parameters
    assert isinstance(instance, CIMInstance)
    assert isinstance(path, CIMInstanceName)

    if instance.classname.lower() != path.classname.lower():
        raise AssertionError(
            _format("Server {0} at {1}: Inconsistent instance and instance "
                    "path: Instance classname {2!A} does not match classname "
                    "of instance path {3!A}",
                    conn.server_definition.nickname, conn.url,
                    instance.classname, path.to_wbem_uri()))

    for key_name in path.keybindings:

        if key_name not in instance.properties:
            raise AssertionError(
                _format("Server {0} at {1}: Inconsistent instance and "
                        "instance path: Instance does not have key property "
                        "{2!A} of instance path {3!A}",
                        conn.server_definition.nickname, conn.url,
                        key_name, path.to_wbem_uri()))

        if instance.properties[key_name].value != \
                path.keybindings[key_name]:
            raise AssertionError(
                _format("Server {0} at {1}: Inconsistent instance and "
                        "instance path: For key {2!A}, instance property "
                        "value {3!A} does not match instance path keybinding "
                        "value {4!A}",
                        conn.server_definition.nickname, conn.url,
                        key_name, instance.properties[key_name],
                        path.keybindings[key_name]))


def assert_mandatory_properties(conn, instance, property_list):
    """
    Assert that an instance has non-null values for a set of properties.

    Parameters:

        conn (WBEMConnection with 'server_definition' attribute)

        instance (CIMInstance): The CIM instance to be verified.

        property_list (iterable of string): The property names.
    """

    # Check parameters
    assert isinstance(instance, CIMInstance)

    instance_prop_names = instance.properties.keys()
    for prop_name in property_list:
        assert isinstance(prop_name, six.string_types)

        if prop_name not in instance_prop_names:
            raise AssertionError(
                _format("Server {0} at {1}: Mandatory properties issue: "
                        "Instance of class {2!A} does not have mandatory "
                        "property {3!A}",
                        conn.server_definition.nickname, conn.url,
                        instance.classname, prop_name))

        prop_value = instance.properties[prop_name]
        if prop_value is None:
            raise AssertionError(
                _format("Server {0} at {1}: Mandatory properties issue: "
                        "Instance of class {2!A} has mandatory property "
                        "{3!A} but with a value of NULL",
                        conn.server_definition.nickname, conn.url,
                        instance.classname, prop_name))


def assert_property_one_of(conn, instance, prop_name, value_list):
    """
    Assert that a simple (= non-array) CIM property of an instance has a
    value that is one of a set of allowable values.

    Parameters:

        conn (WBEMConnection with 'server_definition' attribute)

        instance (CIMInstance): The CIM instance that has the property to
          be verified.

        prop_name (string): Name of the CIM property to be verified.

        value_list (iterable of values): The set of allowable values for
          the property.
    """

    # Check parameters
    assert isinstance(instance, CIMInstance)
    assert isinstance(prop_name, six.string_types)
    prop = instance.properties[prop_name]
    assert not prop.is_array

    prop_value = prop.value
    if prop_value not in value_list:
        raise AssertionError(
            _format("Server {0} at {1}: Property value issue: The value of "
                    "simple property {2!A} in an instance of class {3!A} is "
                    "not in the allowable set of values {4!A}, but is {5!A}",
                    conn.server_definition.nickname, conn.url,
                    prop_name, instance.classname, value_list, prop_value))


def assert_property_contains(conn, instance, prop_name, value):
    """
    Assert that a CIM array property (of an instance) contains a particular
    value.

    Parameters:

        conn (WBEMConnection with 'server_definition' attribute)

        instance (CIMInstance): The CIM instance that has the property to
          be verified.

        prop_name (string): Name of the CIM property to be verified.

        value (value): The value.
    """

    # Check parameters
    assert isinstance(instance, CIMInstance)
    assert isinstance(prop_name, six.string_types)
    prop = instance.properties[prop_name]
    assert prop.is_array

    prop_values = prop.value
    if value not in prop_values:
        raise AssertionError(
            _format("Server {0} at {1}: Property value issue: The value of "
                    "array property {2!A} in an instance of class {3!A} does "
                    "not contain value {4!A}, but is {5!A}",
                    conn.server_definition.nickname, conn.url,
                    prop_name, instance.classname, value, prop_values))


def assert_path_equal(conn, path1, path1_msg, path2, path2_msg):
    """
    Assert that two instance paths are equal, with special treatment
    of their host component:

    - The host component is compared literally (but case insensitively),
      i.e. no IP address to hostname translation is performed.
    - If one or both of the instance paths have their host component set to
      None, the comparison is considered equal.

    For keys of the instance paths that are references, the special
    treatment of their host component is applied recursively.

    Parameters:

        conn (WBEMConnection with 'server_definition' attribute)

        path1 (CIMInstanceName): First instance path to be compared.

        path1_msg (string): Short definition what path1 is.

        path2 (CIMInstanceName): Second instance path to be compared.

        path2_msg (string): Short definition what path2 is.
    """

    # Check parameters
    # Note: path1 and path2 are checked in path_equal()
    assert isinstance(path1_msg, six.string_types)
    assert isinstance(path2_msg, six.string_types)

    if not path_equal(path1, path2):
        raise AssertionError(
            _format("Server {0} at {1}: Instance path issue: Instance path "
                    "{2!A} ({3}) does not match instance path {4!A} ({5})",
                    conn.server_definition.nickname, conn.url,
                    path1.to_wbem_uri(), path1_msg,
                    path2.to_wbem_uri(), path2_msg))


def assert_path_in(conn, path, path_msg, path_list, path_list_msg):
    """
    Assert that an instance path is in a list of instance paths or
    instances, with special treatment of their host component as described
    in assert_path_equal().

    Parameters:

        conn (WBEMConnection with 'server_definition' attribute)

        path (CIMInstanceName): Instance path to be tested for being contained
          in list.

        path_msg (string): Short definition what path is.

        path_list (iterable of CIMInstanceName): List of instance paths tested
          for containing path.

        path_list_msg (string): Short definition what path_list is.
    """

    # Check parameters
    # Note: path and path_list are checked in path_in()
    assert isinstance(path_msg, six.string_types)
    assert isinstance(path_list_msg, six.string_types)

    if not path_in(path, path_list):
        raise AssertionError(
            _format("Server {0} at {1}: Instance path issue: Instance path "
                    "{2!A} ({3}) is not in expected set of instance paths "
                    "({4})",
                    conn.server_definition.nickname, conn.url,
                    path.to_wbem_uri(), path_msg, path_list_msg))


def assert_association_a1(
        conn, profile_id,
        source_path, source_role, assoc_class, far_role, far_class):
    """
    Assert that navigation from a source instance across an association
    succeeds.

    This test performs approach a1:
      - associations: References with manual far end filtering
      - far ends: Associators with operation-based far end filtering

    Parameters:

        conn (WBEMConnection with 'server_definition' attribute)

        profile_id (string): Profile org and name as 'org:name'

        source_path (CIMInstanceName): Instance path of the source instance.

        source_role (string): Filter: Role on source end of association.

        source_class (string): Filter: Class name of source end of association.

        far_role (string): Filter: Role on far end of association.

        far_class (string): Filter: Class name of far end of association.
    """

    a1_assoc_insts = conn.References(
        source_path,
        ResultClass=assoc_class,
        Role=source_role,
        asserted=True)

    a1_assoc_insts = [
        inst for inst in a1_assoc_insts
        if far_role in inst.path.keybindings and
        instance_of(conn, inst.path.keybindings[far_role], far_class)
    ]
    a1_assoc_paths = [inst.path for inst in a1_assoc_insts]

    a1_far_insts = conn.Associators(
        source_path,
        AssocClass=assoc_class,
        ResultClass=far_class,
        Role=source_role,
        ResultRole=far_role,
        asserted=True)

    a1_far_paths = [inst.path for inst in a1_far_insts]

    _assert_association_consistency(
        conn, profile_id,
        source_path, source_role,
        a1_assoc_insts,
        "manually filtered result of References (a1)",
        a1_assoc_paths,
        "paths of manually filtered result of References (a1)",
        assoc_class,
        a1_far_insts,
        "result of Associators (a1)",
        a1_far_paths,
        "paths of result of Associators (a1)",
        far_class, far_role)

    return a1_far_insts, a1_assoc_insts


def assert_association_a2(
        conn, profile_id,
        source_path, source_role, assoc_class, far_role, far_class):
    """
    Assert that navigation from a source instance across an association
    succeeds.

    This test performs approach a2:
      - associations: References with manual far end filtering
      - far end: Associators with manual far end filtering

    Parameters:

        conn (WBEMConnection with 'server_definition' attribute)

        profile_id (string): Profile org and name as 'org:name'

        source_path (CIMInstanceName): Instance path of the source instance.

        source_role (string): Filter: Role on source end of association.

        source_class (string): Filter: Class name of source end of association.

        far_role (string): Filter: Role on far end of association.

        far_class (string): Filter: Class name of far end of association.
    """

    a2_assoc_insts = conn.References(
        source_path,
        ResultClass=assoc_class,
        Role=source_role,
        asserted=True)

    a2_assoc_insts = [
        inst for inst in a2_assoc_insts
        if far_role in inst.path.keybindings and
        instance_of(conn, inst.path.keybindings[far_role], far_class)
    ]
    a2_assoc_paths = [inst.path for inst in a2_assoc_insts]
    a2_assoc_far_paths = [path.keybindings[far_role]
                          for path in a2_assoc_paths]

    a2_far_insts = conn.Associators(
        source_path,
        AssocClass=assoc_class,
        Role=source_role,
        asserted=True)

    a2_far_insts = [
        inst for inst in a2_far_insts
        if path_in(inst.path, a2_assoc_far_paths) and
        instance_of(conn, inst.path, far_class)
    ]
    a2_far_paths = [inst.path for inst in a2_far_insts]

    _assert_association_consistency(
        conn, profile_id,
        source_path, source_role,
        a2_assoc_insts,
        "manually filtered result of References (a2)",
        a2_assoc_paths,
        "paths of manually filtered result of References (a2)",
        assoc_class,
        a2_far_insts,
        "manually filtered result of Associators (a2)",
        a2_far_paths,
        "paths of manually filtered result of Associators (a2)",
        far_class, far_role)

    return a2_far_insts, a2_assoc_insts


def assert_association_a3(
        conn, profile_id,
        source_path, source_role, assoc_class, far_role, far_class):
    """
    Assert that navigation from a source instance across an association
    succeeds.

    This test performs approach a3:
      - associations: ReferenceNames with manual far end filtering and
        GetInstance.
      - far end: AssociatorNames with operation-based far end filtering and
        GetInstance.

    Parameters:

        conn (WBEMConnection with 'server_definition' attribute)

        profile_id (string): Profile org and name as 'org:name'

        source_path (CIMInstanceName): Instance path of the source instance.

        source_role (string): Filter: Role on source end of association.

        source_class (string): Filter: Class name of source end of association.

        far_role (string): Filter: Role on far end of association.

        far_class (string): Filter: Class name of far end of association.
    """

    a3_assoc_paths = conn.ReferenceNames(
        source_path,
        ResultClass=assoc_class,
        Role=source_role,
        asserted=True)

    a3_assoc_paths = [
        path for path in a3_assoc_paths
        if far_role in path.keybindings and
        instance_of(conn, path.keybindings[far_role], far_class)
    ]
    a3_assoc_insts = []

    for path in a3_assoc_paths:
        _inst = conn.GetInstance(path, asserted=True)
        a3_assoc_insts.append(_inst)

    a3_far_paths = conn.AssociatorNames(
        source_path,
        AssocClass=assoc_class,
        ResultClass=far_class,
        Role=source_role,
        ResultRole=far_role,
        asserted=True)

    a3_far_insts = []
    for path in a3_far_paths:
        _inst = conn.GetInstance(path, asserted=True)
        a3_far_insts.append(_inst)

    _assert_association_consistency(
        conn, profile_id,
        source_path, source_role,
        a3_assoc_insts,
        "GetInstance on manually filtered result of ReferenceNames (a3)",
        a3_assoc_paths,
        "manually filtered result of ReferenceNames (a3)",
        assoc_class,
        a3_far_insts,
        "GetInstance on result of AssociatorNames (a3)",
        a3_far_paths,
        "result of AssociatorNames (a3)",
        far_class, far_role)

    return a3_far_insts, a3_assoc_insts


def assert_association_a4(
        conn, profile_id,
        source_path, source_role, assoc_class, far_role, far_class):
    """
    Assert that navigation from a source instance across an association
    succeeds.

    This test performs approach a4:
      - associations: ReferenceNames with manual far end filtering and
        GetInstance.
      - far end: AssociatorNames with manual far end filtering and
        GetInstance.

    Parameters:

        conn (WBEMConnection with 'server_definition' attribute)

        profile_id (string): Profile org and name as 'org:name'

        source_path (CIMInstanceName): Instance path of the source instance.

        source_role (string): Filter: Role on source end of association.

        source_class (string): Filter: Class name of source end of association.

        far_role (string): Filter: Role on far end of association.

        far_class (string): Filter: Class name of far end of association.
    """

    a4_assoc_paths = conn.ReferenceNames(
        source_path,
        ResultClass=assoc_class,
        Role=source_role,
        asserted=True)

    a4_assoc_paths = [
        path for path in a4_assoc_paths
        if far_role in path.keybindings and
        instance_of(conn, path.keybindings[far_role], far_class)
    ]
    a4_assoc_far_paths = [path.keybindings[far_role]
                          for path in a4_assoc_paths]

    a4_assoc_insts = []
    for path in a4_assoc_paths:
        _inst = conn.GetInstance(path, asserted=True)
        a4_assoc_insts.append(_inst)

    a4_far_paths = conn.AssociatorNames(
        source_path,
        AssocClass=assoc_class,
        Role=source_role,
        asserted=True)

    a4_far_paths = [
        path for path in a4_far_paths
        if path_in(path, a4_assoc_far_paths) and
        instance_of(conn, path, far_class)
    ]

    a4_far_insts = []
    for path in a4_far_paths:
        _inst = conn.GetInstance(path, asserted=True)
        a4_far_insts.append(_inst)

    _assert_association_consistency(
        conn, profile_id,
        source_path, source_role,
        a4_assoc_insts,
        "GetInstance on manually filtered result of ReferenceNames (a4)",
        a4_assoc_paths,
        "manually filtered result of ReferenceNames (a4)",
        assoc_class,
        a4_far_insts,
        "GetInstance on manually filtered result of AssociatorNames (a4)",
        a4_far_paths,
        "manually filtered result of AssociatorNames (a4)",
        far_class, far_role)

    return a4_far_insts, a4_assoc_insts


def assert_association_a5(
        conn, profile_id,
        source_path, source_role, assoc_class, far_role, far_class):
    """
    Assert that navigation from a source instance across an association
    succeeds.

    This test performs approach a5:
      - associations: EnumerateInstances of the association and manual
        filtering.
      - far end: GetInstance on the far end keys of the associations.

    Parameters:

        conn (WBEMConnection with 'server_definition' attribute)

        profile_id (string): Profile org and name as 'org:name'

        source_path (CIMInstanceName): Instance path of the source instance.

        source_role (string): Filter: Role on source end of association.

        source_class (string): Filter: Class name of source end of association.

        far_role (string): Filter: Role on far end of association.

        far_class (string): Filter: Class name of far end of association.
    """

    a5_assoc_insts = conn.EnumerateInstances(
        namespace=source_path.namespace,
        ClassName=assoc_class,
        asserted=True)

    a5_assoc_insts = [
        inst for inst in a5_assoc_insts
        if source_role in inst.path.keybindings and
        path_equal(inst.path.keybindings[source_role], source_path) and
        far_role in inst.path.keybindings and
        instance_of(conn, inst.path.keybindings[far_role], far_class)
    ]
    a5_assoc_paths = [inst.path for inst in a5_assoc_insts]
    a5_assoc_far_paths = [path.keybindings[far_role]
                          for path in a5_assoc_paths]

    a5_far_insts = []
    for path in a5_assoc_far_paths:
        _inst = conn.GetInstance(path, asserted=True)
        a5_far_insts.append(_inst)

    a5_far_paths = [inst.path for inst in a5_far_insts]

    _assert_association_consistency(
        conn, profile_id,
        source_path, source_role,
        a5_assoc_insts,
        "manually filtered result of "
        "EnumerateInstances on association class (a5)",
        a5_assoc_paths,
        "paths of manually filtered result of "
        "EnumerateInstances on association class (a5)",
        assoc_class,
        a5_far_insts,
        "GetInstance on far ends of manually filtered result of "
        "EnumerateInstances on association class (a5)",
        a5_far_paths,
        "far ends of manually filtered result of "
        "EnumerateInstances on association class (a5)",
        far_class, far_role)

    return a5_far_insts, a5_assoc_insts


def assert_association_a6(
        conn, profile_id,
        source_path, source_role, assoc_class, far_role, far_class):
    """
    Assert that navigation from a source instance across an association
    succeeds.

    This test performs approach a6:
      - associations: EnumerateInstanceNames of the association and manual
        filtering, followed by GetInstance.
      - far end: GetInstance on the far end keys of the associations.

    Parameters:

        conn (WBEMConnection with 'server_definition' attribute)

        profile_id (string): Profile org and name as 'org:name'

        source_path (CIMInstanceName): Instance path of the source instance.

        source_role (string): Filter: Role on source end of association.

        source_class (string): Filter: Class name of source end of association.

        far_role (string): Filter: Role on far end of association.

        far_class (string): Filter: Class name of far end of association.
    """

    a6_assoc_paths = conn.EnumerateInstanceNames(
        namespace=source_path.namespace,
        ClassName=assoc_class,
        asserted=True)

    a6_assoc_paths = [
        path for path in a6_assoc_paths
        if source_role in path.keybindings and
        path_equal(path.keybindings[source_role], source_path) and
        far_role in path.keybindings and
        instance_of(conn, path.keybindings[far_role], far_class)
    ]
    a6_assoc_far_paths = [path.keybindings[far_role]
                          for path in a6_assoc_paths]

    a6_assoc_insts = []
    for path in a6_assoc_paths:
        _inst = conn.GetInstance(path, asserted=True)
        a6_assoc_insts.append(_inst)

    a6_far_insts = []
    for path in a6_assoc_far_paths:
        _inst = conn.GetInstance(path, asserted=True)
        a6_far_insts.append(_inst)

    a6_far_paths = [inst.path for inst in a6_far_insts]

    _assert_association_consistency(
        conn, profile_id,
        source_path, source_role,
        a6_assoc_insts,
        "GetInstance on manually filtered result of "
        "EnumerateInstanceNames on association class (a6)",
        a6_assoc_paths,
        "manually filtered result of "
        "EnumerateInstanceNames on association class (a6)",
        assoc_class,
        a6_far_insts,
        "GetInstance on far ends of manually filtered result of "
        "EnumerateInstanceNames on association class (a6)",
        a6_far_paths,
        "far ends of manually filtered result of "
        "EnumerateInstanceNames on association class (a6)",
        far_class, far_role)

    return a6_far_insts, a6_assoc_insts


def _assert_association_consistency(
        conn, profile_id,
        source_path, source_role,
        assoc_insts, assoc_insts_msg, assoc_paths, assoc_paths_msg,
        assoc_class,
        far_insts, far_insts_msg, far_paths, far_paths_msg,
        far_class, far_role):

    if len(far_insts) != len(far_paths):
        raise AssertionError(
            _format("Server {0} at {1}: Number of far end instances {2} ({3}) "
                    "does not match number of far end paths {4} ({5})",
                    conn.server_definition.nickname, conn.url,
                    len(far_insts), far_insts_msg,
                    len(far_paths), far_paths_msg))

    if len(assoc_insts) != len(assoc_paths):
        raise AssertionError(
            _format("Server {0} at {1}: Number of association instances {2} "
                    "({3}) does not match number of association paths "
                    "{4} ({5})",
                    conn.server_definition.nickname, conn.url,
                    len(assoc_insts), assoc_insts_msg,
                    len(assoc_paths), assoc_paths_msg))

    if len(far_insts) != len(assoc_insts):
        raise AssertionError(
            _format("Server {0} at {1}: Number of far end instances {2} ({3}) "
                    "does not match number of association instances {4} ({5})",
                    conn.server_definition.nickname, conn.url,
                    len(far_insts), far_insts_msg,
                    len(assoc_insts), assoc_insts_msg))

    assert_instance_of(conn, far_insts, far_class)
    for inst in far_insts:
        assert_instance_consistency(conn, inst, inst.path)
    for path in far_paths:
        assert_path_in(
            conn, path, far_paths_msg,
            far_insts, "path of {0}".format(far_insts_msg))

    assert_instance_of(conn, assoc_insts, assoc_class)
    for inst in assoc_insts:
        assert_instance_consistency(conn, inst, inst.path)
        assert source_role in inst.path.keybindings
        assert_path_equal(
            conn,
            inst.path.keybindings[source_role],
            "source end {0!r} of {1}".format(source_role, assoc_insts_msg),
            source_path,
            "source instance")
        assert far_role in inst.path.keybindings
        assert_path_in(
            conn,
            inst.path.keybindings[far_role],
            "far end {0!r} of {1}".format(far_role, assoc_insts_msg),
            far_paths,
            far_paths_msg)
    for path in assoc_paths:
        assert_path_in(
            conn,
            path, assoc_paths_msg,
            assoc_insts, "path of {0}".format(assoc_insts_msg))

    # Check consistency across the association approaches
    # Py test does not seem to have a reliable order of executing the
    # testcases w.r.t. items in fixtures. Therefore, we have to assume
    # that this function is invoked in arbitrary order w.r.t. the
    # association approaches. The algorithn we use is that the
    # first approach defines the results all subsequent approaches
    # compare against.
    prefix = ':'.join([profile_id, source_path.to_wbem_uri(),
                       source_role, assoc_class, far_role, far_class])
    far_paths_id = prefix + ':far_paths'
    assoc_paths_id = prefix + ':assoc_paths'
    if not OBJECT_CACHE.has_list(conn.url, far_paths_id):
        # This is the first approach executed (not necessarily a1).
        # Store the results in the object cache.
        OBJECT_CACHE.add_list(
            conn.url, far_paths_id, far_paths)
        OBJECT_CACHE.add_list(
            conn.url, assoc_paths_id, assoc_paths)
    else:
        # This is not the first approach executed.

        # Retrieve the result of the first approach from the object cache.
        first_far_paths = OBJECT_CACHE.get_list(
            conn.url, far_paths_id)
        first_assoc_paths = OBJECT_CACHE.get_list(
            conn.url, assoc_paths_id)

        # Check consistency with first approach
        assert len(assoc_paths) == len(first_assoc_paths)
        for path in assoc_paths:
            assert_path_in(
                conn,
                path,
                assoc_paths_msg,
                first_assoc_paths,
                assoc_paths_msg + " of first executed association approach")
        assert len(far_paths) == len(first_far_paths)
        for path in far_paths:
            assert_path_in(
                conn,
                path,
                far_paths_msg,
                first_far_paths,
                far_paths_msg + " of first executed association approach")


def std_uri(instance):
    """
    Return canonical WBEM URI of the path of the instance, nullifying the
    host, or the empty string.
    """
    if instance is None or instance.path is None:
        return ''
    else:
        path = instance.path.copy()
        path.host = None
        return path.to_wbem_uri(format='canonical')


def assert_profile_tree(conn, profile_inst, profile_ancestry,
                        reference_direction, tls_org, tls_name):
    """
    Assert that a profile tree is a tree without circular references,
    when navigating to referenced profiles in the specified reference
    direction.

    Parameters:

        conn (WBEMConnection with 'server_definition' attribute)

        profile_inst (CIMInstance): Profile that is tested.

        profile_ancestry (dict): Profile ancestry up to the top level spec,
          as a dict with:
            * key: std_uri of profile that is tested.
            * value: std_uri of its referencing profile (or spec, for the top).

        reference_direction (string): Reference direction to use for the test
          ('dmtf' or 'snia').

        tls_org (string), tls_name (string): Org and name of top level spec
          of the profile tree (used only for failure messages).
    """

    profile_uri = std_uri(profile_inst)

    # from pprint import pprint
    # print("Debug: assert_profile_tree called")
    # print("       profile:  {0}".format(profile_uri))
    # print("       ancestry: ")
    # for k in profile_ancestry.keys():
    #     print("                 {0}".format(k))

    if reference_direction == 'dmtf':
        result_role_down = 'Antecedent'
        # result_role_up = 'Dependent'
    else:
        assert reference_direction == 'snia'
        result_role_down = 'Dependent'
        # result_role_up = 'Antecedent'

    sub_profile_insts = conn.Associators(
        profile_inst.path,
        AssocClass='CIM_ReferencedProfile',
        ResultRole=result_role_down,
        asserted=True)

    if not sub_profile_insts:
        # The profile is a leaf profile, i.e. does not reference any further
        # profiles.
        pass
    else:
        for sub_profile_inst in sub_profile_insts:

            sub_profile_uri = std_uri(sub_profile_inst)

            if sub_profile_uri in profile_ancestry:
                raise AssertionError(
                    "Server {0} at {1}: Profile tree under top level "
                    "specification {2} {3!r} has a circular reference: "
                    "Profile at {4} references profile at {5} which is "
                    "already in its own reference ancestry {6}".
                    format(conn.server_definition.nickname, conn.url,
                           tls_org, tls_name,
                           profile_uri, sub_profile_uri,
                           profile_ancestry.keys()))

            profile_ancestry[sub_profile_uri] = profile_uri

            assert_profile_tree(
                conn, sub_profile_inst, profile_ancestry,
                reference_direction, tls_org, tls_name)

            del profile_ancestry[sub_profile_uri]



#
# (C) Copyright 2004,2005 Hewlett-Packard Development Company, L.P.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# Author: Tim Potter <tpot@hp.com>
#

"""
Utility function for validation of XML strings against the CIM-XML DTD.

The CIM-XML DTD is defined in DSP0201 as a text standard that shows the DTD
in the text. DSP0203 is the complete DTD, but is provided only for convenience.
In case of differences, DSP0201 overrules DSP0203.

Note: The CIM_DTD_V22.dtd file used for DTD checking by pywbem for some time
      is a preliminary version that contained a few elements that were never
      released as a DMTF standard:
        TABLE, TABLEROW, TABLEROW.DECLARATION, TABLECELL.DECLARATION,
        TABLECELL.REFERENCE, RESPONSEDESTINATION.

Note: Changes in DSP0203 2.3.1, compared to 2.2.0 (final):
      - SIMPLERSP: Removed child SIMPLEREQACK, to fix a an incompletely removed
        left-over from the preliminary version of 2.2.
      - RETURNVALUE: Made its child elements optional, in support of void
        CIM methods.
      - IMETHODRESPONSE: Added PARAMVALUE*, in support of pull operations. Note
        this was inconsistent and IPARAMVALUE should have been used.
      - PARAMVALUE: Added choices for children CLASSNAME, CLASS, INSTANCE,
        VALUE.NAMEDINSTANCE, in support of pull operations (became necessary
        as a result of using PARAMVALUE in IMETHODRESPONSE).
      - ENUMERATIONCONTEXT: Added this element in support of pull operations.

Note: Changes in DSP0203 2.4.0, compared to 2.3.1:
      - KEYVALUE: Its TYPE attribute is now required. (incompatible!)
      - SIMPLEREQ: Added optional CORRELATOR* children, in support of operation
        correlation.
      - CORRELATOR: Added this element in support of operation correlation.
      - PARAMVALUE: Added choice for children INSTANCENAME* in support of
        pull operations (became necessary as a result of using PARAMVALUE in
        IMETHODRESPONSE).
      - IRETURNVALUE: Added choice for children VALUE.INSTANCEWITHPATH* in
        support of pull operations.
      - ENUMERATIONCONTEXT: Removed this element again, because representation
        of enumeration context value in pull operations was changed to string
        in DSP0200.
      - EXPPARAMVALUE: Removed incorrect children VALUE, METHODRESPONSE,
        IMETHODRESPONSE, because they were not needed.
"""

from __future__ import print_function, absolute_import

import os
import os.path
import re
from subprocess import Popen, PIPE, STDOUT

from pywbem._utils import _ensure_bytes, _ensure_unicode

# CIM-XML DTD file to validate against
DTD_FILE = os.path.join('tests', 'dtd', 'DSP0203_2.3.1.dtd')

# TODO: The DTD 2.4.0 requires the TYPE attribute on the KEYVALUE elements,
#       and this causes some tests in test_cim_xml.py to fail. Fix them.


class CIMXMLValidationError(Exception):
    """
    Exception indicating a CIM-XML validation error.
    """
    pass


def validate_cim_xml(cim_xml_str, root_elem_name=None):
    """
    Validate a CIM-XML string against the CIM-XML DTD, optionally
    requiring a particular XML root element.

    If the validation succeeds, the function returns. Otherwise,
    `CIMXMLValidationError` is raised and its exception message is the
    (possible multi-line) output of the `xmllint` command.

    Parameters:

      cim_xml_str (string): CIM-XML string to be validated.

      root_elem_name (string): Name of XML element that is expected as the root
        element in the CIM-XML string to be validated.
        `None` means no checking for a particular root element is performed.

    Raises:

      CIMXMLValidationError: CIM-XML validation error
    """

    # The DOCTYPE instruction needs the DTD file with forward slashes.
    # Also, the xmllint used on Windows requires forward slashes and complains
    # with "Could not parse DTD tests\dtd\DSP0203_2.3.1.dtd" if invoked
    # with backslashes.
    dtd_file_fw = DTD_FILE.replace('\\', '/')

    # Make sure the validator checks the specified root element, if any
    if root_elem_name is not None:
        cim_xml_str = u'<!DOCTYPE {0} SYSTEM "{1}">\n{2}'. \
            format(root_elem_name, dtd_file_fw, cim_xml_str)
        xmllint_cmd = 'xmllint --valid --noout -'
    else:
        xmllint_cmd = 'xmllint --dtdvalid {0} --noout -'.format(dtd_file_fw)

    p = Popen(xmllint_cmd, stdout=PIPE, stderr=STDOUT, stdin=PIPE, shell=True)

    cim_xml_str = _ensure_bytes(cim_xml_str)
    p.stdin.write(cim_xml_str)
    p.stdin.close()

    status = p.wait()
    if status != 0:
        out_lines = p.stdout.readlines()
        output = _ensure_unicode(b'\n'.join(out_lines))
        raise CIMXMLValidationError(output)


def validate_cim_xml_obj(obj, obj_xml_str, exp_xml_str):
    """
    Validate a CIM-XML string of a CIM object against an expected CIM-XML
    string and against the CIM-XML DTD.

    Parameters:

      obj (CIM object): The CIM object that belongs to the CIM-XML string,
       just for reporting purposes.

      obj_xml_str (string): The CIM-XML string to be validated.

      exp_xml_str (string): The expected CIM-XML string.
    """

    assert obj_xml_str == exp_xml_str

    m = re.match(r'^<([^ >]+)', exp_xml_str)
    assert m is not None, \
        "Testcase issue: Cannot determine root element from expected " \
        "CIM-XML string:\n{0}".format(exp_xml_str)
    root_elem_name = m.group(1)

    try:
        validate_cim_xml(obj_xml_str, root_elem_name)
    except CIMXMLValidationError as exc:
        raise AssertionError(
            "DTD validation of CIM-XML for {0} object failed:\n"
            "{1}\n"
            "Required XML root element: {2}\n"
            "CIM-XML string:\n"
            "{3}".
            format(type(obj), exc, root_elem_name, obj_xml_str))
